import json
from typing import Optional

from hbreader import hbread
from rdflib import Graph


from linkml_runtime.dumpers.dumper_root import Dumper
from linkml_runtime.utils.context_utils import CONTEXTS_PARAM_TYPE, CONTEXT_TYPE
from linkml_runtime.utils.formatutils import remove_empty_items
from linkml_runtime.utils.yamlutils import YAMLRoot


class RDFDumper(Dumper):
    def as_rdf_graph(self, element: YAMLRoot, contexts: CONTEXTS_PARAM_TYPE, namespaces: CONTEXT_TYPE = None) -> Graph:
        """
        Convert element into an RDF graph guided by the context(s) in contexts
        :param element: element to represent in RDF
        :param contexts: JSON-LD context(s) in the form of:
            * file name
            * URL
            * JSON String
            * dict
            * JSON Object
            * A list containing elements of any type named above
        :param namespaces: A file name, URL, JSON String, dict or JSON object that includes the set of namespaces to
        be bound to the return graph.  If absent, contexts get used
        :return: rdflib Graph containing element
        """
        if contexts is None:
            raise Exception(f'Must pass in JSON-LD context via contexts parameter')
        if isinstance(contexts, list):
            inp_contexts = [json.loads(hbread(c)) for c in contexts]
        else:
            inp_contexts = json.loads(hbread(contexts))

        from linkml_runtime.dumpers import json_dumper
        jsonld_str = json_dumper.dumps(element)
        g = Graph().parse(data=jsonld_str, format='json-ld')
        #rdf_jsonld = expand()
        #g = rdflib_graph_from_pyld_jsonld(rdf_jsonld)

        if namespaces is not None:
            ns_source = json.loads(hbread(namespaces))
        else:
            ns_source = inp_contexts

        # TODO: make a utility out of this or add it to prefixcommons
        if ns_source and '@context' in ns_source:
            ns_contexts = ns_source['@context']
            if isinstance(ns_contexts, dict):
                ns_contexts = [ns_contexts]
            for ns_context in ns_contexts:
                if isinstance(ns_context, dict):
                    for pfx, ns in ns_context.items():
                        if isinstance(ns, dict):
                            if '@id' in ns and ns.get('@prefix', False):
                                ns = ns['@id']
                            else:
                                continue
                        if not pfx.startswith('@'):
                            g.bind(pfx, ns)

        return g

    def dump(self, element: YAMLRoot, to_file: str, contexts: CONTEXTS_PARAM_TYPE = None, fmt: str = 'turtle') -> None:
        """
        Write element as rdf to to_file
        :param element: LinkML object to be emitted
        :param to_file: file to write to
        :param contexts: JSON-LD context(s) in the form of:
            * file name
            * URL
            * JSON String
            * dict
            * JSON Object
            * A list containing elements of any type named above
        :param fmt: RDF format
        """
        super().dump(element, to_file, contexts=contexts, fmt=fmt)

    def dumps(self, element: YAMLRoot, contexts: CONTEXTS_PARAM_TYPE = None, fmt: Optional[str] = 'turtle') -> str:
        """
        Convert element into an RDF graph guided by the context(s) in contexts
        :param element: element to represent in RDF
        :param contexts: JSON-LD context(s) in the form of a file or URL, a json string or a json obj
        :param fmt: rdf format
        :return: rdflib Graph containing element
        """
        return self.as_rdf_graph(remove_empty_items(element, hide_protected_keys=True), contexts).\
            serialize(format=fmt)



from dataclasses import dataclass
from typing import Dict, Any, Optional, ClassVar, List, Union

from rdflib import URIRef

from linkml_runtime.utils.curienamespace import CurieNamespace
from linkml_runtime.utils.metamodelcore import URIorCURIE, empty_list, URI
from linkml_runtime.utils.yamlutils import YAMLRoot, extended_str
from utils.enumerations import EnumDefinitionImpl

LINKML = CurieNamespace('linkml', 'https://w3id.org/linkml/')

RenderingURI = URI

class PermissibleValueText(extended_str):
    pass

@dataclass
class PermissibleValue(YAMLRoot):
    """
    a permissible value, accompanied by intended text and an optional mapping to a concept URI
    """
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = LINKML.PermissibleValue
    class_class_curie: ClassVar[str] = "linkml:PermissibleValue"
    class_name: ClassVar[str] = "permissible_value"
    class_model_uri: ClassVar[URIRef] = LINKML.PermissibleValue

    text: Union[str, PermissibleValueText] = None
    description: Optional[str] = None
    meaning: Optional[Union[str, URIorCURIE]] = None
    deprecated: Optional[str] = None
    todos: Optional[Union[str, List[str]]] = empty_list()
    notes: Optional[Union[str, List[str]]] = empty_list()
    comments: Optional[Union[str, List[str]]] = empty_list()
    from_schema: Optional[Union[str, URI]] = None
    imported_from: Optional[str] = None
    see_also: Optional[Union[Union[str, URIorCURIE], List[Union[str, URIorCURIE]]]] = empty_list()
    deprecated_element_has_exact_replacement: Optional[Union[str, URIorCURIE]] = None
    deprecated_element_has_possible_replacement: Optional[Union[str, URIorCURIE]] = None
    is_a: Optional[Union[str, PermissibleValueText]] = None
    mixins: Optional[Union[Union[str, PermissibleValueText], List[Union[str, PermissibleValueText]]]] = empty_list()

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self.text is None:
            raise ValueError("text must be supplied")
        if not isinstance(self.text, PermissibleValueText):
            self.text = PermissibleValueText(self.text)

        if self.description is not None and not isinstance(self.description, str):
            self.description = str(self.description)

        if self.meaning is not None and not isinstance(self.meaning, URIorCURIE):
            self.meaning = URIorCURIE(self.meaning)

        if self.deprecated is not None and not isinstance(self.deprecated, str):
            self.deprecated = str(self.deprecated)

        if self.todos is None:
            self.todos = []
        if not isinstance(self.todos, list):
            self.todos = [self.todos]
        self.todos = [v if isinstance(v, str) else str(v) for v in self.todos]

        if self.notes is None:
            self.notes = []
        if not isinstance(self.notes, list):
            self.notes = [self.notes]
        self.notes = [v if isinstance(v, str) else str(v) for v in self.notes]

        if self.comments is None:
            self.comments = []
        if not isinstance(self.comments, list):
            self.comments = [self.comments]
        self.comments = [v if isinstance(v, str) else str(v) for v in self.comments]

        if self.from_schema is not None and not isinstance(self.from_schema, URI):
            self.from_schema = URI(self.from_schema)

        if self.imported_from is not None and not isinstance(self.imported_from, str):
            self.imported_from = str(self.imported_from)

        if self.see_also is None:
            self.see_also = []
        if not isinstance(self.see_also, list):
            self.see_also = [self.see_also]
        self.see_also = [v if isinstance(v, URIorCURIE) else URIorCURIE(v) for v in self.see_also]

        if self.deprecated_element_has_exact_replacement is not None and not isinstance(self.deprecated_element_has_exact_replacement, URIorCURIE):
            self.deprecated_element_has_exact_replacement = URIorCURIE(self.deprecated_element_has_exact_replacement)

        if self.deprecated_element_has_possible_replacement is not None and not isinstance(self.deprecated_element_has_possible_replacement, URIorCURIE):
            self.deprecated_element_has_possible_replacement = URIorCURIE(self.deprecated_element_has_possible_replacement)

        if self.is_a is not None and not isinstance(self.is_a, PermissibleValueText):
            self.is_a = PermissibleValueText(self.is_a)

        if self.mixins is None:
            self.mixins = []
        if not isinstance(self.mixins, list):
            self.mixins = [self.mixins]
        self.mixins = [v if isinstance(v, PermissibleValueText) else PermissibleValueText(v) for v in self.mixins]

        super().__post_init__(**kwargs)


# Enumerations
class PvFormulaOptions(EnumDefinitionImpl):
    """
    The formula used to generate the set of permissible values from the code_set values
    """
    CODE = PermissibleValue(text="CODE",
                               description="The permissible values are the set of possible codes in the code set")
    CURIE = PermissibleValue(text="CURIE",
                                 description="The permissible values are the set of CURIES in the code set")
    URI = PermissibleValue(text="URI",
                             description="The permissible values are the set of code URIs in the code set")
    FHIR_CODING = PermissibleValue(text="FHIR_CODING",
                                             description="The permissible values are the set of FHIR coding elements derived from the code set")

    _defn = EnumDefinition(
        name="PvFormulaOptions",
        description="The formula used to generate the set of permissible values from the code_set values",
    )

class PermissibleValueImpl(PermissibleValue):
    """
    Permissible Value implementation
    """
    def __init__(self, *args, defn: EnumDefinitionImpl, **kwargs) -> None:
        """ Record the referencing definition to allow the entry to be fleshed out from a terminology service """
        super().__init__(*args, **kwargs)
        self._defn = defn

    def __post_init__(self, **kwargs: Dict[str, Any]) -> None:
        """ Make sure that we are correctly situated in the containing definition """
        if self.text in self._defn:
            if self._defn.permissible_values[self.text] != self:
                raise TypeError(f"Permissible value for code: {self.text} is already assigned")
        if self not in self._defn.permissible_values:
            if not self._defn.code_set:
                # Fixed set of values -- no terminology reference
                raise TypeError("{self}: Permissible value not in definition set")



import re
from decimal import Decimal
from numbers import Number
from typing import List, Any, Union

from jsonasobj2 import JsonObj, as_dict, is_list, is_dict, items, as_json_obj

ws_pattern = re.compile(r'\s+')
us_pattern = re.compile(r'_+')



def camelcase(txt: str) -> str:
    def _up(s: str):
        return s[0].upper() + (s[1:] if len(s) > 1 else '')

    return ''.join([_up(word) for word in us_pattern.sub(' ', txt.strip().replace(',', '')).split()])


def underscore(txt: str) -> str:
    return ws_pattern.sub('_', txt.strip()).replace(',', '').replace('-', '_')


def lcamelcase(txt: str) -> str:
    s = camelcase(txt)
    return s[0].lower() + s[1:]


def be(entry: object) -> str:
    """ Return a stringified version of object replacing Nones with empty strings """
    return str(entry).strip() if entry else ''


def mangled_attribute_name(clsname: str, attributename: str) -> str:
    """ Return the mangling we use for attributes definitions """
    return lcamelcase(clsname) + '__' + underscore(attributename)


split_col = 115


def sfx(uri: str) -> str:
    """
    Add a separator to a uri if none exists.

    Note: This should only be used with module id's -- it is not uncommon to use partial prefixes,
    e.g. PREFIX bfo: http://purl.obolibrary.org/obo/BFO_
    :param uri: uri to be suffixed
    :return: URI with suffix
    """
    return str(uri) + ('' if uri.endswith(('/', '#', '_', ':')) else '/')


def uri_for(prefix: str, suffix: str) -> str:
    """ Generator for predicate and identifier URI's """
    if ':' in prefix:
        return sfx(prefix) + suffix
    else:
        return prefix + ':' + suffix


def split_line(txt: str, split_len: int = split_col) -> List[str]:
    # TODO: consider replacing by textwrap.fill function, but note that its behavior is a bit different
    out_lines = []
    words = txt.split()
    cur_line = ""
    for word in words:
        word += ' '
        if len(cur_line) + len(word) > split_len:
            out_lines.append(cur_line if cur_line else word)
            if not cur_line:
                word = ""
            else:
                cur_line = ""
        cur_line += word
    if cur_line:
        out_lines.append(cur_line)
    return out_lines


def wrapped_annotation(txt: str) -> str:
    rval = []
    for line in [line.strip() for line in txt.split('\n')]:
        if len(line) > split_col:
            rval += split_line(line)
        else:
            rval.append(line)
    return '\n\t'.join(rval)

def shex_results_as_string(rslts) -> str:
    """ Pretty print ShEx Evaluation result """
    # TODO: Add this method to ShEx itself
    rval = [f"Evalutating: {str(rslts.focus)} against {str(rslts.start)}"]
    if rslts.result:
        rval.append("Result: CONFORMS")
    else:
        rval.append("Result: NonConforming")
    rval += rslts.reason.split('\n')
    return '\n'.join(rval)


def is_empty(v: Any) -> bool:
    """
    Determine whether v is considered "empty" in the LinkML context.
    An element is "empty" if:
    1) it is None
    2) It is an empty dictionary
    3) It is an empty list
    4) It is an empty JsonObj
    """
    return v is None or (isinstance(v, (dict, list)) and not v) or (isinstance(v, JsonObj) and not as_dict(v))


def remove_empty_items(obj: Any, hide_protected_keys: bool = False, inside: bool = False) -> Any:
    """
    Recursively iterate over obj removing any empty internal entries.  Note:  this returns a _copy_ of obj of we are
    dealing with a list or a dictionary.

    If hide_protected_keys is true, any key that begins with an underscore is removed from the structure, meaning that
       {
          '_k1': {
             'x': ...
          }
       }
       becomes
       {
          'x': ...
       }

    The above situation ONLY applies when there is ONE k,v pair and v is a dictionary

    Note that this will also convert Decimals to floats or ints; this is necessary
    as both json dumpers and yaml dumpers will encode Decimal types by default.
    See https://github.com/linkml/linkml/issues
    
    This is easier than fixing the individual serializers, described here:
    
    - JSON: https://bugs.python.org/issue16535, https://stackoverflow.com/questions/1960516/python-json-serialize-a-decimal-object
    - YAML: https://stackoverflow.com/questions/21695705/dump-an-python-object-as-yaml-file/51261042, https://github.com/yaml/pyyaml/pull/372

    :param obj: Object to be tweaked
    :param hide_protected_keys: True means remove keys that begin with an underscore
    :param inside: Keep from removing the outermost container
    :return: copy of obj with empty items removed or None if obj itself is "empty"
    """
    if is_list(obj):
        # for discussion of logic, see: https://github.com/linkml/linkml-runtime/issues/42
        obj_list = [e for e in [remove_empty_items(l, hide_protected_keys=hide_protected_keys, inside=True)
                                for l in obj if l != '_root'] if not is_empty(e)]
        return obj_list if not inside or not is_empty(obj_list) else None
    elif is_dict(obj):
        obj_dict = {k: v for k, v in [(k2, remove_empty_items(v2, hide_protected_keys=hide_protected_keys, inside=True))
                                      for k2, v2 in items(obj)] if not is_empty(v)}

        # https://github.com/linkml/linkml/issues/119
        # Remove the additional level of nesting with enums
        if len(obj_dict) == 1 and list(obj_dict.keys())[0] == '_code':
            enum_text = list(obj_dict.values())[0].get('text', None)
            if enum_text is not None:
                return enum_text
        if hide_protected_keys and len(obj_dict) == 1 and str(list(obj_dict.keys())[0]).startswith('_'):
            inner_element = list(obj_dict.values())[0]
            if isinstance(inner_element, dict):
                obj_dict = inner_element
        return obj_dict if not inside or not is_empty(obj_dict) else None
    elif is_empty(obj):
        return None
    elif isinstance(obj, Decimal):
        # note that attempting to implement https://bugs.python.org/issue16535
        # will not work for yaml serializations
        v = str(obj)
        if '.' in v and not v.endswith('.0'):
            return float(obj)
        else:
            return int(obj)
    else:
        return obj



from io import StringIO
from typing import Union, TextIO, Optional, Dict, Type, List

import yaml
from hbreader import FileInfo

from linkml_runtime.loaders.loader_root import Loader
from linkml_runtime.utils.yamlutils import YAMLRoot, DupCheckYamlLoader


class YAMLLoader(Loader):
    """
    A Loader that is capable of instantiating LinkML data objects from a YAML file
    """

    def load_any(self, source: Union[str, dict, TextIO], target_class: Type[YAMLRoot], *, base_dir: Optional[str] = None,
                 metadata: Optional[FileInfo] = None, **_) -> Union[YAMLRoot, List[YAMLRoot]]:
        def loader(data: Union[str, dict], _: FileInfo) -> Optional[Dict]:
            return yaml.load(StringIO(data), DupCheckYamlLoader) if isinstance(data, str) else data

        if not metadata:
            metadata = FileInfo()
        if base_dir and not metadata.base_path:
            metadata.base_path = base_dir
        return self.load_source(source, loader, target_class, accept_header="text/yaml, application/yaml;q=0.9",
                                metadata=metadata)

    def loads_any(self, source: str, target_class: Type[YAMLRoot], *, metadata: Optional[FileInfo] = None, **_) -> Union[YAMLRoot, List[YAMLRoot]]:
        """
        Load source as a string
        @param source: source
        @param target_class: destination class
        @param metadata: metadata about the source
        @param _: extensions
        @return: instance of taarget_class
        """
        return self.load_any(source, target_class, metadata=metadata)


from .types import String, Integer, Boolean, Float, Double, Decimal, Time, Date, Datetime, Uriorcurie, Uri, \
    Ncname, Objectidentifier, Nodeidentifier
from .extensions import Extension, Extensible
from .annotations import Annotation, Annotatable
from .meta import ElementName, SchemaDefinitionName, TypeDefinitionName, SubsetDefinitionName, DefinitionName, \
    EnumDefinitionName, SlotDefinitionName, ClassDefinitionName, PrefixPrefixPrefix, LocalNameLocalNameSource, \
    AltDescriptionSource, PermissibleValueText, Element, SchemaDefinition, TypeDefinition, SubsetDefinition, \
    Definition, EnumDefinition, SlotDefinition, ClassDefinition, Prefix, LocalName, Example, AltDescription, \
    PermissibleValue, PvFormulaOptions




"""
Packaging for working with LinkML distributions
"""
import logging
import pkgutil
from pathlib import PurePath
from typing import List, Type



def get_default_paths(file_type: str) -> List[PurePath]:
    """
    Return candidate relative paths for a file type

    :param file_type:
    :return: candidate paths
    """
    paths = []
    rel_dirs = []
    srcp = PurePath('src')
    if file_type == 'yaml':
        rel_dirs = [PurePath('model') /'schema',
                    PurePath('schema'),
                    srcp / 'linkml',
                    srcp / 'model',
                    srcp / 'model' / 'schema',
                    srcp / 'schema',
                    ]
    elif file_type == 'schema.json':
        rel_dirs = [PurePath('jsonschema')]
    elif file_type == 'context.jsonld':
        rel_dirs = [PurePath('jsonld')]
    elif '.' in file_type:
        # e.g. 'owl.ttl' --> 'owl'
        rel_dirs = [PurePath(file_type.split('.')[0])]
    else:
        rel_dirs = []
    for rel_dir in rel_dirs:
        paths.append(rel_dir)
    paths.append(PurePath('.'))
    return paths

def get_packaged_file_as_str(package: str, file_type: str, rel_paths: List[PurePath]=[], encoding="utf-8") -> str:
    """
    Retrieve the value of a data file distributed alongside a python package

    :param package: Python package as string, e.g. linkml_runtime.linkml_model.meta
    :param file_type: e.g. 'yaml', 'schema.json', 'owl.ttl'
    :param rel_path: optional relative path from python package to
    :param encoding:
    :return:
    """
    parts = package.split('.')
    package_name = parts[-1]
    suffix = file_type
    data = None
    for path in get_default_paths(file_type) + rel_paths:
        try:
            full_path = path / f'{package_name}.{suffix}'
            data = pkgutil.get_data(package, str(full_path))
            break
        except FileNotFoundError:
            logging.debug(f'candidate {path} not found')
    if not data:
        raise FileNotFoundError(f'package: {package} file: {file_type}')
    return data.decode(encoding)

def get_schema_string(package: str, **kwargs) -> str:
    return get_packaged_file_as_str(package, file_type='yaml', **kwargs)

def get_jsonschema_string(package: str, **kwargs) -> str:
    return get_packaged_file_as_str(package, file_type='schema.json', **kwargs)



from typing import Union, TextIO, Optional, Type, List

from hbreader import FileInfo

from linkml_runtime.loaders.loader_root import Loader
from linkml_runtime.utils.context_utils import CONTEXTS_PARAM_TYPE
from linkml_runtime.utils.yamlutils import YAMLRoot
from rdflib import Graph

from linkml_runtime.loaders.requests_ssl_patch import no_ssl_verification

# TODO: figure out what mime types go here.  I think we can find the complete set in rdflib
RDF_MIME_TYPES = "application/x-turtle;q=0.9, application/rdf+n3;q=0.8, application/rdf+xml;q=0.5, text/plain;q=0.1"


class RDFLoader(Loader):

    def load_any(self, *args, **kwargs) -> Union[YAMLRoot, List[YAMLRoot]]:
        return self.load(*args, **kwargs)


    def load(self, source: Union[str, TextIO, Graph], target_class: Type[YAMLRoot], *, base_dir: Optional[str] = None,
             contexts: CONTEXTS_PARAM_TYPE = None, fmt: Optional[str] = 'turtle',
             metadata: Optional[FileInfo] = None) -> YAMLRoot:
        """
        Load the RDF in source into the python target_class structure
        :param source: RDF data source. Can be a URL, a file name, an RDF string, an open handle or an existing graph
        :param base_dir: Base directory that can be used if file name or URL.  This is copied into metadata if present
        :param target_class: LinkML class to load the RDF into
        :param contexts: JSON-LD context(s) to use to generate the JSON that will be loaded into target_class.  This is
        optional because, if source is in JSON-LD format, it is possible that the contexts are already there
        :param fmt: format of source if it isn't an existing Graph
        :param metadata: source information. Used by some loaders to record where information came from
        :return: Instance of target_class
        """

        def loader(data: Union[str, dict], _: FileInfo) -> Optional[dict]:
            """
            Process an RDF graph or a JSON-LD string.  We do this by using pyld_jsonld_from_rdflib_graph to
            emit a JSON-LD string and then process it with jsonld.frame.

            :param data: Graph or JSON-LD string
            :param _: Unused - part of signature for other implementations
            :return: Dictionary to load into the target class
            """
            # Where we sit -- we don't have the start string (and other info?) in the context, so we don't get a package
            # on the way out
            # Also - if we don't pop type below, we need to act accordingly
            if isinstance(data, str):
                if fmt != 'json-ld':
                    g = Graph()
                    g.parse(data=data, format=fmt)
                    jsonld_str = g.serialize(format='json-ld', indent=4)
                    data = json.loads(jsonld_str)
                    #data = pyld_jsonld_from_rdflib_graph(g)

            if not isinstance(data, dict):
                # TODO: Add a context processor to the source w/ CONTEXTS_PARAM_TYPE
                # TODO: figure out what to do base options below
                # TODO: determine whether jsonld.frame can handle something other than string input
                # TODO: see https://github.com/RDFLib/rdflib/issues/1727
                # frame = {'@context': contexts, '@type': f'{target_class.__name__}'}
                #data_as_dict = jsonld.frame(data, contexts)
                data_as_dict = data
            else:
                data_as_dict = data
            typ = data_as_dict.pop('@type', None)
            # TODO: remove this when we get the Biolinkml issue fixed
            if not typ:
                typ = data_as_dict.pop('type', None)
            if typ and typ != target_class.class_name:
                # TODO: connect this up with the logging facility or warning?
                print(f"Warning: input type mismatch. Expected: {target_class.__name__}, Actual: {typ}")
            return self.json_clean(data_as_dict)

        if not metadata:
            metadata = FileInfo()
        if base_dir and not metadata.base_path:
            metadata.base_path = base_dir

        # If the input is a graph, convert it to JSON-LD
        if isinstance(source, Graph):
            source = pyld_jsonld_from_rdflib_graph(source)
            jsonld_str = source.serialize(format='json-ld', indent=4)
            source = json.loads(jsonld_str)
            fmt = 'json-ld'

        # While we may want to allow full SSL verification at some point, the general philosophy is that content forgery
        # is not going to be a serious problem.
        # TODO: Make the SSL option a settable parameter in the package itself
        with no_ssl_verification():
            return self.load_source(source, loader, target_class, accept_header=RDF_MIME_TYPES, metadata=metadata)



from copy import copy
from json import JSONDecoder
from typing import Union, Any, List, Optional, Type, Callable, Dict

import yaml
from deprecated.classic import deprecated
from jsonasobj2 import JsonObj, as_json, as_dict, JsonObjTypes, items
import jsonasobj2
from rdflib import Graph
from yaml.constructor import ConstructorError

from linkml_runtime.utils.context_utils import CONTEXTS_PARAM_TYPE, merge_contexts
from linkml_runtime.utils.formatutils import is_empty

YAMLObjTypes = Union[JsonObjTypes, "YAMLRoot"]


class YAMLMark(yaml.error.Mark):
    def __str__(self):
        snippet = self.get_snippet()
        where = "\nFile \"%s\", line %d, column %d"   \
                % (self.name, self.line+1, self.column+1)
        if snippet is not None:
            where += ":\n"+snippet
        return where

class YAMLRoot(JsonObj):
    """
    The root object for all python YAML representations
    """
    def __init__(self, *args, **kwargs):
        """
        Override dataclass initializer
        @param args:
        @param kwargs:
        """
        super().__init__(*args, **kwargs)

    def __post_init__(self, *args: List[str],  **kwargs):
        if args or kwargs:
            messages: List[str] = []
            for v in args:
                v = repr(v)[:40].replace('\n', '\\n')
                messages.append(f"Unknown positional argument: {v}")
            for k in kwargs.keys():
                v = repr(kwargs[k])[:40].replace('\n', '\\n')
                messages.append(f"{TypedNode.yaml_loc(k)} Unknown argument: {k} = {v}")
            raise ValueError('\n'.join(messages))

    def _default(self, obj, filtr: Callable[[dict], dict] = None):
        """ JSON serializer callback.
        1) Filter out empty values (None, {}, [] and False) and mangle the names
        2) Add ID entries for dictionary entries

        :param obj: YAMLRoot object to serialize
        :param filtr: Filter to remove elements
        :return: Serialized version of obj
        """

        if isinstance(obj, YAMLRoot):
            rval = dict()
            for k, v in (filtr(obj.__dict__) if filtr else obj.__dict__).items():
                is_classvar = k.startswith("type_") and hasattr(type(obj), k)
                if is_classvar:
                    print(f"***** {k} is classvar ")
                if not is_classvar and not k.startswith('_') and v is not None and\
                        (not isinstance(v, (dict, list, bool)) or v):

                    from linkml_runtime.utils.enumerations import EnumDefinitionImpl
                    if isinstance(v, dict):
                        itemslist = []
                        for vk, vv in v.items():
                            # if isinstance(vv, ClassDefinition):
                            #     vv['@id'] = camelcase(vk)
                            # elif isinstance(vv, (SlotDefinition, TypeDefinition)):
                            #     if k != 'slot_usage':
                            #         vv['@id'] = underscore(vk)
                            itemslist.append(vv)
                        rval[k] = itemslist
                    # TODO: Figure out how to make EnumDefinitionImpl a subclass of EnumDefinition
                    # elif isinstance(v, EnumDefinition):
                    elif isinstance(v, EnumDefinitionImpl):
                        rval[k] = v.code
                    else:
                        rval[k] = v
            return rval
        else:
            return obj._default(obj) if hasattr(obj, '_default') and callable(obj._default) else\
                JSONDecoder().decode(obj)

    @staticmethod
    def _is_empty(v: Any) -> bool:
        # TODO: Deprecate this function and migrate the python generator over to the stand alone is_empty
        return is_empty(v)

    def _normalize_inlined_as_list(self, slot_name: str, slot_type: Type, key_name: str, keyed: bool) -> None:
        self._normalize_inlined(slot_name, slot_type, key_name, keyed, True)

    def _normalize_inlined_as_dict(self, slot_name: str, slot_type: Type, key_name: str, keyed: bool) -> None:
        self._normalize_inlined(slot_name, slot_type, key_name, keyed, False)

    def _normalize_inlined(self, slot_name: str, slot_type: Type, key_name: str, keyed: bool, is_list: bool) \
            -> None:
        """
         __post_init__ function for a list of inlined keyed or identified classes.

        The input to this is either a list or dictionary of dictionaries.  In the list case, every key entry
        in the list must be unique.  In the dictionary case, the key may or may not be repeated in the dictionary
        element. The internal storage structure is a dictionary of dictionaries.
        @param slot_name: Name of the slot being normalized
        @param slot_type: Slot range type
        @param key_name: Name of the key or identifier in the range
        @param keyed: True means each identifier must be unique
        @param is_list: True means inlined as list
        """
        raw_slot: Union[list, dict, JsonObj] = self[slot_name]
        if raw_slot is None:
            raw_slot = []
        elif not isinstance(raw_slot, (dict, list, JsonObj)):
            raw_slot = [raw_slot]
        cooked_slot = list() if is_list else dict()
        cooked_keys = set()

        def order_up(key: Any, cooked_entry: YAMLRoot) -> None:
            """ A cooked entry is ready to be added to the return slot """
            if cooked_entry[key_name] != key:
                raise ValueError(
                    f"Slot: {loc(slot_name)} - attribute {loc(key_name)} " \
                    f"value ({loc(cooked_entry[key_name])}) does not match key ({loc(key)})")
            if keyed and key in cooked_keys:
                raise ValueError(f"{loc(key)}: duplicate key")
            cooked_keys.add(key)
            if is_list:
                cooked_slot.append(cooked_entry)
            else:
                cooked_slot[key] = cooked_entry

        def loc(s):
            loc_str = TypedNode.yaml_loc(s) if isinstance(s, TypedNode) else ''
            if loc_str == ': ':
                loc_str = ''
            return loc_str + str(s)

        def form_1(entries: Dict[Any, Optional[Union[dict, JsonObj]]]) -> None:
            """ A dictionary of key:dict entries where key is the identifier and dict is an instance of slot_type """
            for key, raw_obj in items(entries):
                if raw_obj is None:
                    raw_obj = {}
                if key_name not in raw_obj:
                    raw_obj = copy(raw_obj)
                    raw_obj[key_name] = key
                if not issubclass(type(raw_obj), slot_type):
                    order_up(key, slot_type(**as_dict(raw_obj)))
                else:
                    order_up(key, raw_obj)

        # TODO: Make an external function extract a root JSON list
        if isinstance(raw_slot, JsonObj):
            raw_slot = raw_slot._hide_list()

        if isinstance(raw_slot, list):
            # We have a list of entries
            for list_entry in raw_slot:
                if isinstance(list_entry, (dict, JsonObj)):
                    # list_entry is either a key:dict, key_name:value or **kwargs
                    if len(list_entry) == 1:
                        # key:dict or key_name:key
                        for lek, lev in items(list_entry):
                            if lek == key_name and not isinstance(lev, (list, dict, JsonObj)):
                                # key_name:value
                                order_up(list_entry[lek], slot_type(list_entry))
                                break   # Not strictly necessary, but
                            elif not isinstance(lev, (list, dict, JsonObj)):
                                # key: value --> slot_type(key, value)
                                order_up(lek, slot_type(lek, lev))
                            else:
                                form_1(list_entry)
                    else:
                        # **kwargs
                        cooked_obj = slot_type(**as_dict(list_entry))
                        order_up(cooked_obj[key_name], cooked_obj)
                elif isinstance(list_entry, list):
                    # *args
                    cooked_obj = slot_type(*list_entry)
                    order_up(cooked_obj[key_name], cooked_obj)
                else:
                    # lone key [key1: , key2: ... }
                    order_up(list_entry, slot_type(**{key_name: list_entry}))
        else:
            # We have a dictionary
            if key_name in raw_slot and not isinstance(raw_slot[key_name], (list, dict, JsonObj)):
                # Vanilla dictionary - {key: v11, s12: v12, ...}
                order_up(raw_slot[key_name], slot_type(**as_dict(raw_slot)))
            else:
                # We have either {key1: {obj1}, key2: {obj2}...} or {key1:, key2:, ...}
                for k, v in items(raw_slot):
                    if v is None:
                        v = dict()
                    if isinstance(v, (dict, JsonObj)):
                        form_1({k: v})
                    elif not isinstance(v, list):
                        order_up(k, slot_type(*[k, v]))
                    else:
                        raise ValueError(f"Unrecognized entry: {loc(k)}: {str(v)}")
        self[slot_name] = cooked_slot

    def _normalize_inlined_slot(self, slot_name: str, slot_type: Type, key_name: Optional[str],
                                inlined_as_list: Optional[bool], keyed: bool) -> None:
        """
        A deprecated entry point to slot normalization. Used for models generated prior to the linkml-runtime split.

        This code is invoked via one of the following patterns:
            if self.<slot> is None:
                self.<slot> = []
            if not isinstance(self.<slot>, list):
                self.extensions = [self.<slot>]
            self._normalize_inlined_slot(slot_name="<slot>", slot_type=<type>, key_name="<key>", inlined_as_list=True, keyed=...)
        or
            if self.<slot> is None:
                self.<slot> = []
            if not isinstance(self.<slot>, (list, dict)):
                self.local_names = [self.<slot>]
            self._normalize_inlined_slot(slot_name="<slot>", slot_type=<type>, key_name="<key>", inlined_as_list=None, keyed=...)

        The above pattern broke when the new jsonasobj was introduced, which is why we have the normalization above.  The code
        below reverse engineers the above and invokes the new form

        """
        if inlined_as_list:
            list_slot = self[slot_name]
            if len(list_slot) == 1 and not isinstance(list_slot[0], list):
                self[slot_name] = list_slot[0]
            self._normalize_inlined_as_list(slot_name, slot_type, key_name, keyed)
        else:
            dict_slot = self[slot_name]
            if isinstance(dict_slot, list):
                if len(dict_slot) == 1 and not isinstance(dict_slot[0], (list, dict)):
                    self[slot_name] = dict_slot[0]
            self._normalize_inlined_as_dict(slot_name, slot_type, key_name, keyed)

    # ==================
    # Error intercepts
    # ==================
    def MissingRequiredField(self, field_name: str) -> None:
        """ Generic loader error handler """
        raise ValueError(f"{field_name} must be supplied")


def root_representer(dumper: yaml.Dumper, data: YAMLRoot):
    """ YAML callback -- used to filter out empty values (None, {}, [] and false)

    @param dumper: data dumper
    @param data: data to be dumped
    @return:
    """
    # TODO: Figure out how to import EnumDefinition here
    # elif isinstance(v, EnumDefinition):
    from linkml_runtime.utils.enumerations import EnumDefinitionImpl
    if isinstance(data, EnumDefinitionImpl):
        data = data.code
    rval = dict()
    for k, v in data.__dict__.items():
        if not k.startswith('_') and v is not None and (not isinstance(v, (dict, list)) or v):
            rval[k] = v
    return dumper.represent_data(rval)


def from_yaml(data: str, cls: Type[YAMLRoot]) -> YAMLRoot:
    return cls(**yaml.load(data, DupCheckYamlLoader))


def as_yaml(element: YAMLRoot) -> str:
    """
    Return element in a YAML representation

    Uses SafeDumper, key-vals are omitted if val is None

    :param element: YAML object
    :return: Stringified representation
    """
    return yaml.dump(element, Dumper=yaml.SafeDumper, sort_keys=False)


def as_json_object(element: YAMLRoot, contexts: CONTEXTS_PARAM_TYPE = None, inject_type = True) -> JsonObj:
    """
    Return the representation of element as a JsonObj object
    :param element: element to return
    :param contexts: context(s) to include in the output
    :param inject_type: if True (default), add a @type at the top level
    :return: JsonObj representation of element
    """
    rval = copy(element)
    if inject_type:
        rval['@type'] = element.__class__.__name__
    context_element = merge_contexts(contexts)
    if context_element:
        rval['@context'] = context_element['@context']
    return rval


class TypedNode:
    def __init__(self, v: Union[Any, "TypedNode"]):
        self._s = v._s if isinstance(v, TypedNode) else None
        self._len = v._len if isinstance(v, TypedNode) else None
        super().__init__()

    def add_node(self, node):
        self._s = node.start_mark
        self._len = node.end_mark.index - node.start_mark.index
        return self

    @deprecated(reason="Use yaml_loc instead")
    def loc(self) -> str:
        return self._loc()

    def _loc(self) -> str:
        return f'File "{self._s.name}", line {self._s.line + 1}, col {self._s.column + 1}' if self._s else ''

    @staticmethod
    def yaml_loc(loc_str: Optional[Union["TypedNode", str]] = None, suffix: Optional[str] = ": ") -> str:
        """ Return the yaml file and location of loc_str if it exists """
        return '' if loc_str is None or not hasattr(loc_str, "_loc" or not callable(loc_str._loc)) else\
            (loc_str._loc() + suffix)


class extended_str(str, TypedNode):
    def concat(self, *items) -> "extended_str":
        rval = extended_str(str(self) + ''.join([str(item) for item in items]))
        for item in items[::-1]:
            if isinstance(item, TypedNode):
                rval._s = item._s
                rval._len = item._len
                break
        return rval


class extended_int(int, TypedNode):
    pass


class extended_float(float, TypedNode):
    pass


class DupCheckYamlLoader(yaml.loader.SafeLoader):
    """
    A YAML loader that throws an error when the same key appears twice
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.add_constructor(yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG, DupCheckYamlLoader.map_constructor)
        self.add_constructor(yaml.resolver.BaseResolver.DEFAULT_SEQUENCE_TAG, DupCheckYamlLoader.seq_constructor)
        self.add_constructor('tag:yaml.org,2002:str', DupCheckYamlLoader.construct_yaml_str)
        self.add_constructor('tag:yaml.org,2002:int', DupCheckYamlLoader.construct_yaml_int)
        self.add_constructor('tag:yaml.org,2002:float', DupCheckYamlLoader.construct_yaml_float)

    def get_mark(self):
        if self.stream is None:
            return YAMLMark(self.name, self.index, self.line, self.column, self.buffer, self.pointer)
        else:
            return YAMLMark(self.name, self.index, self.line, self.column, None, None)

    def construct_yaml_int(self, node):
        """ Scalar constructor that returns the node information as the value """
        return extended_int(super().construct_yaml_int(node)).add_node(node)

    def construct_yaml_str(self, node):
        """ Scalar constructor that returns the node information as the value """
        return extended_str(super().construct_yaml_str(node)).add_node(node)

    def construct_yaml_float(self, node):
        """ Scalar constructor that returns the node information as the value """
        return extended_float(super().construct_yaml_float(node)).add_node(node)

    @staticmethod
    def map_constructor(loader,  node, deep=False):
        """ Duplicate of constructor.construct_mapping w/ exception that we check for dups

        """
        if not isinstance(node, yaml.MappingNode):
            raise ConstructorError(None, None, "expected a mapping node, but found %s" % node.id, node.start_mark)
        mapping = {}
        for key_node, value_node in node.value:
            key = loader.construct_object(key_node, deep=deep)
            value = loader.construct_object(value_node, deep=deep)
            if key in mapping:
                raise ValueError(f"Duplicate key: \"{key}\"")
            mapping[key] = value
        return mapping

    @staticmethod
    def seq_constructor(loader, node, deep=False):
        if not isinstance(node, yaml.SequenceNode):
            raise ConstructorError(None, None,
                                   "expected a sequence node, but found %s" % node.id,
                                   node.start_mark)
        for child in node.value:
            if not child.value:
                raise ConstructorError(None, None, "Empty list elements are not allowed", node.start_mark)
        return [loader.construct_object(child, deep=deep)
                for child in node.value]


yaml.SafeDumper.add_multi_representer(YAMLRoot, root_representer)
yaml.SafeDumper.add_multi_representer(extended_str, yaml.SafeDumper.represent_str)
yaml.SafeDumper.add_multi_representer(extended_int, yaml.SafeDumper.represent_int)
yaml.SafeDumper.add_multi_representer(extended_float, yaml.SafeDumper.represent_float)
yaml.SafeDumper.add_multi_representer(str, yaml.SafeDumper.represent_str)
yaml.SafeDumper.add_multi_representer(int, yaml.SafeDumper.represent_int)
yaml.SafeDumper.add_multi_representer(float, yaml.SafeDumper.represent_float)


def as_rdf(element: YAMLRoot, contexts: CONTEXTS_PARAM_TYPE = None) -> Graph:
    """
    Convert element into an RDF graph guided by the context(s) in contexts
    :param element: element to represent in RDF
    :param contexts: JSON-LD context(s) in the form of a file or URL name, a json string or a json obj
    :return: rdflib Graph containing element
    """

    jsonld = as_json_object(element, contexts)
    graph = Graph()
    graph.parse(data=as_json(jsonld), format="json-ld", prefix=True)
    return graph



from decimal import Decimal

import yaml

from linkml_runtime.dumpers.dumper_root import Dumper
from linkml_runtime.utils.formatutils import remove_empty_items
from linkml_runtime.utils.yamlutils import YAMLRoot

class YAMLDumper(Dumper):

    def dumps(self, element: YAMLRoot, **kwargs) -> str:
        """ Return element formatted as a YAML string """
        # Internal note: remove_empty_items will also convert Decimals to int/float;
        # this is necessary until https://github.com/yaml/pyyaml/pull/372 is merged
        return yaml.dump(remove_empty_items(element, hide_protected_keys=True),
                         Dumper=yaml.SafeDumper, sort_keys=False, 
                         allow_unicode=True,
                         **kwargs)



from json_flattener import unflatten_from_csv, KeyConfig, GlobalConfig, Serializer
import json
from typing import Type, Union, List
from linkml_runtime.utils.yamlutils import YAMLRoot

from linkml_runtime.loaders.loader_root import Loader
from linkml_runtime.loaders.json_loader import JSONLoader
from linkml_runtime.linkml_model.meta import SlotDefinitionName, SchemaDefinition, ClassDefinition
from linkml_runtime.utils.yamlutils import YAMLRoot
from linkml_runtime.utils.schemaview import SchemaView
from linkml_runtime.utils.csvutils import get_configmap

class CSVLoader(Loader):

    def load_any(self, *args, **kwargs) -> Union[YAMLRoot, List[YAMLRoot]]:
        return self.load(*args, **kwargs)


    def loads(self, input,
              target_class: Type[YAMLRoot],
              index_slot: SlotDefinitionName = None,
              schema: SchemaDefinition = None,
              schemaview: SchemaView = None,
              **kwargs) -> str:
        if schemaview is None:
            schemaview = SchemaView(schema)
        configmap = get_configmap(schemaview, index_slot)
        config = GlobalConfig(key_configs=configmap)
        objs = unflatten_from_csv(input, config=config, **kwargs)
        return JSONLoader().loads(json.dumps({index_slot: objs}), target_class=target_class)

    def load(self, source: str,
             target_class: Type[YAMLRoot],
             index_slot: SlotDefinitionName = None,
             schema: SchemaDefinition = None,
             schemaview: SchemaView = None,
             **kwargs) -> str:
        if schemaview is None:
            schemaview = SchemaView(schema)
        configmap = get_configmap(schemaview, index_slot)
        config = GlobalConfig(key_configs=configmap)
        print(f'Loading from {source}')
        objs = unflatten_from_csv(source, config=config, **kwargs)
        return JSONLoader().loads(json.dumps({index_slot: objs}), target_class=target_class)


import warnings
import contextlib

import requests
from urllib3.exceptions import InsecureRequestWarning

# ***** Taken from https://stackoverflow.com/questions/15445981/how-do-i-disable-the-security-certificate-check-in-python-requests
old_merge_environment_settings = requests.Session.merge_environment_settings


@contextlib.contextmanager
def no_ssl_verification():
    opened_adapters = set()

    def merge_environment_settings(self, url, proxies, stream, verify, cert):
        # Verification happens only once per connection so we need to close
        # all the opened adapters once we're done. Otherwise, the effects of
        # verify=False persist beyond the end of this context manager.
        opened_adapters.add(self.get_adapter(url))

        settings = old_merge_environment_settings(self, url, proxies, stream, verify, cert)
        settings['verify'] = False

        return settings

    requests.Session.merge_environment_settings = merge_environment_settings

    try:
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', InsecureRequestWarning)
            yield
    finally:
        requests.Session.merge_environment_settings = old_merge_environment_settings

        for adapter in opened_adapters:
            try:
                adapter.close()
            except Exception:
                pass



from linkml_runtime.utils.curienamespace import CurieNamespace
from linkml_runtime.utils.schemaview import SchemaView
from rdflib import RDF, RDFS, SKOS, XSD, OWL

# use importlib.metadata to read the version provided
# by the package during installation. Do not hardcode
# the version in the code
try:
    import importlib.metadata as importlib_metadata
except ModuleNotFoundError:
    import importlib_metadata

LINKML = CurieNamespace('linkml', 'https://w3id.org/linkml/')
TCCM = CurieNamespace('tccm', 'https://ontologies.r.us/tccm/')
OWL = CurieNamespace('owl', OWL)
RDF = CurieNamespace('rdf', RDF)
RDFS = CurieNamespace('rdfs', RDFS)
SKOS = CurieNamespace('skos', SKOS)
XSD = CurieNamespace('xsd', XSD)

__version__ = importlib_metadata.version(__name__)


class MappingError(ValueError):
    """
    An error when mapping elements of a LinkML model to runtime objects
    """
    pass

class DataNotFoundError(ValueError):
    """
    An error in which data cannot be found
    """
    pass



from linkml_runtime.loaders.json_loader import JSONLoader
from linkml_runtime.loaders.rdf_loader import RDFLoader
from linkml_runtime.loaders.rdflib_loader import RDFLibLoader
from linkml_runtime.loaders.yaml_loader import YAMLLoader
from linkml_runtime.loaders.csv_loader import CSVLoader

json_loader = JSONLoader()
rdf_loader = RDFLoader()
rdflib_loader = RDFLibLoader()
yaml_loader = YAMLLoader()
csv_loader = CSVLoader()



import io
import yaml
import json
from typing import Dict, List, Any

from linkml_runtime.dumpers.dumper_root import Dumper
from linkml_runtime.dumpers.json_dumper import JSONDumper
from linkml_runtime.utils.yamlutils import YAMLRoot
from linkml_runtime.linkml_model.meta import SlotDefinitionName, SchemaDefinition
from linkml_runtime.utils.schemaview import SchemaView

from linkml_runtime.utils.csvutils import GlobalConfig, get_configmap
from json_flattener import flatten_to_csv


class CSVDumper(Dumper):

    def dumps(self, element: YAMLRoot,
              index_slot: SlotDefinitionName = None,
              schema: SchemaDefinition = None,
              schemaview: SchemaView = None,
              **kwargs) -> str:
        """ Return element formatted as CSV lines """
        json_dumper = JSONDumper()
        element_j = json.loads(json_dumper.dumps(element))
        objs = element_j[index_slot]
        if schemaview is None:
            schemaview = SchemaView(schema)
        configmap = get_configmap(schemaview, index_slot)
        config = GlobalConfig(key_configs=configmap)
        output = io.StringIO()
        flatten_to_csv(objs, output, config=config, **kwargs)
        return output.getvalue()



from abc import ABC, abstractmethod

from linkml_runtime.utils.yamlutils import YAMLRoot


class Dumper(ABC):
    """ Abstract base class for all dumpers """

    def dump(self, element: YAMLRoot, to_file: str, **_) -> None:
        """
        Write element to to_file
        :param element: LinkML object to be dumped
        :param to_file: file to dump to
        :@param _: method specific arguments
        """
        with open(to_file, 'w', encoding='UTF-8') as output_file:
            output_file.write(self.dumps(element, **_))

    @abstractmethod
    def dumps(self, element: YAMLRoot, **_) -> str:
        """
        Convert element to a string
        @param element: YAMLRoot object to be rendered
        @param _: method specific arguments
        @return: stringified representation of element
        """
        raise NotImplementedError()



import json
import os
from typing import List, Optional, Union


def flatten_dict(ctxt: str, base: str, seen: Optional[List[str]] = None) -> dict:

    def map_context(ctxt_ent: Union[str, dict, list], seen: List[str]) -> Union[dict, list]:
        if isinstance(ctxt_ent, str):
            ent_dict = flatten_dict(ctxt_ent, base, seen)
            return ent_dict['@context'] if '@context' in ent_dict else ent_dict
        elif isinstance(ctxt_ent, list):
            return [map_context(clent, seen) for clent in ctxt_ent]
        else:
            return map_dict(ctxt_ent, seen)

    def map_dict(inp: dict, seen: List[str]) -> dict:
        rval = dict()
        for k, v in inp.items():
            if k == '@context':
                v = map_context(v, seen)
            elif k == '@import':
                v = {}
            elif isinstance(v, dict):
                v = map_dict(v, seen)
            if v:
                rval[k] = v
        return rval

    if seen is None:
        seen = []
    elif ctxt in seen:
        return {}
    seen.append(ctxt)
    with open(os.path.join(base, ctxt)) as f:
        rval = map_dict(json.load(f), seen)
    seen.pop()
    return rval


def flatten(ctxt: str, base: str) -> str:
    print('_'*10 + f' Flattening {os.path.join(base, ctxt)}')
    return json.dumps(flatten_dict(ctxt, base), indent=2)


# ctxt_file = flatten('Package.context.jsonld', LD_11_DIR)
# with open(os.path.join(LD_11_DIR, 'termci_schema_inlined.context.jsonld'), 'w') as outf:
#     outf.write(ctxt_file)



import json
import logging
from typing import Union, TextIO, Optional, Dict, Type, List

from hbreader import FileInfo

from linkml_runtime.loaders.loader_root import Loader
from linkml_runtime.utils.yamlutils import YAMLRoot


class JSONLoader(Loader):

    def load_any(self, source: Union[str, dict, TextIO], target_class: Type[YAMLRoot], *, base_dir: Optional[str] = None,
             metadata: Optional[FileInfo] = None, **_) -> Union[YAMLRoot, List[YAMLRoot]]:
        def loader(data: Union[str, dict], _: FileInfo) -> Optional[Dict]:
            data_as_dict = json.loads(data) if isinstance(data, str) else data
            if isinstance(data_as_dict, list):
                return self.json_clean(data_as_dict)
            typ = data_as_dict.pop('@type', None)
            if typ and typ != target_class.__name__:
                logging.warning(f"Warning: input type mismatch. Expected: {target_class.__name__}, Actual: {typ}")
            return self.json_clean(data_as_dict)

        if not metadata:
            metadata = FileInfo()
        if base_dir and not metadata.base_path:
            metadata.base_path = base_dir
        return self.load_source(source, loader, target_class,
                                accept_header="application/ld+json, application/json, text/json", metadata=metadata)



from linkml_runtime.dumpers.json_dumper import JSONDumper
from linkml_runtime.dumpers.rdf_dumper import RDFDumper
from linkml_runtime.dumpers.rdflib_dumper import RDFLibDumper
from linkml_runtime.dumpers.yaml_dumper import YAMLDumper
from linkml_runtime.dumpers.csv_dumper import CSVDumper

json_dumper = JSONDumper()
rdf_dumper = RDFDumper()
rdflib_dumper = RDFLibDumper()
yaml_dumper = YAMLDumper()
csv_dumper = CSVDumper()



import os
from enum import Enum, auto
from typing import Optional, Union

import requests
from rdflib import Namespace

LINKML_URL_BASE = "https://w3id.org/linkml/"
LINKML_NAMESPACE = Namespace(LINKML_URL_BASE)
GITHUB_IO_BASE = "https://linkml.github.io/linkml-model/"
GITHUB_BASE = "https://raw.githubusercontent.com/linkml/linkml-model/"
LOCAL_BASE = os.path.abspath(os.path.dirname(__file__))
GITHUB_API_BASE = "https://api.github.com/repos/linkml/linkml-model/"
GITHUB_RELEASES = GITHUB_BASE + "releases"
GITHUB_TAGS = GITHUB_BASE + "tags"




class _AutoName(Enum):
    @staticmethod
    def _generate_next_value_(name, start, count, last_values):
        return name.lower()


class Source(_AutoName):
    """ LinkML package source name """
    META = auto()
    TYPES = auto()
    MAPPINGS = auto()
    ANNOTATIONS = auto()
    EXTENSIONS = auto()


class Format(Enum):
    """ LinkML package formats """
    GRAPHQL = "graphql"
    HTML = ""
    JSON = "json"
    JSONLD = "context.jsonld"
    JSON_SCHEMA = "schema.json"
    NATIVE_JSONLD = "model.context.jsonld"
    NATIVE_RDF = "model.ttl"
    NATIVE_SHEXC = "model.shex"
    NATIVE_SHEXJ = "model.shexj"
    OWL = "owl.ttl"
    PYTHON = "py"
    RDF = "ttl"
    SHEXC = "shex"
    SHEXJ = "shexj"
    YAML = "yaml"


class _Path(Enum):
    """ LinkML Relative paths"""
    GRAPHQL = "graphql"
    HTML = "docs"
    JSON = "json"
    JSONLD = "jsonld"
    JSON_SCHEMA = "jsonschema"
    NATIVE_JSONLD = "jsonld"
    NATIVE_RDF = "ttl"
    NATIVE_SHEXC = "shex"
    NATIVE_SHEXJ = "shex"
    OWL = "owl"
    PYTHON = "linkml_model"
    RDF = "rdf"
    SHEXC = "shex"
    SHEXJ = "shex"
    YAML = "model/schema"


class ReleaseTag(_AutoName):
    """ Release tags
    LATEST - the absolute latest in the supplied branch
    CURRENT - the latest _released_ version in the supplied branch """
    LATEST = auto()
    CURRENT = auto()


def _build_path(source: Source, fmt: Format) -> str:
    """ Create the relative path for source and fmt """
    return f"{_Path[fmt.name].value}/{source.value}.{fmt.value}"


def _build_loc(base: str, source: Source, fmt: Format) -> str:
    return f"{base}{_build_path(source, fmt)}".replace('blob/', '')


def URL_FOR(source: Source, fmt: Format) -> str:
    """ Return the URL to retrieve source in format """
    return f"{LINKML_URL_BASE}{source.value}.{fmt.value}"


def LOCAL_PATH_FOR(source: Source, fmt: Format) -> str:
    return os.path.join(LOCAL_BASE, _build_path(source, fmt))


def GITHUB_IO_PATH_FOR(source: Source, fmt: Format) -> str:
    return _build_loc(GITHUB_IO_BASE, source, fmt)


def GITHUB_PATH_FOR(source: Source,
                    fmt: Format,
                    release: Optional[Union[ReleaseTag, str]] = ReleaseTag.CURRENT,
                    branch: Optional[str] = "main") -> str:
    def do_request(url) -> object:
        resp = requests.get(url)
        if resp.ok:
            return resp.json()
        raise requests.HTTPError(f"{resp.status_code} - {resp.reason}: {url}")

    def tag_to_commit(tag: str) -> str:
        tags = do_request(f"{GITHUB_API_BASE}tags")
        for tagent in tags:
            if tagent['name'] == tag:
                return _build_loc(f"{GITHUB_BASE}blob/{tagent['commit']['sha']}/", source, fmt)
        raise ValueError(f"Tag: {tag} not found!")

    if release is not ReleaseTag.CURRENT and branch != "main":
        raise ValueError("Cannot specify both a release and a branch")

    # Return the absolute latest entry for branch
    if release is ReleaseTag.LATEST or (release is ReleaseTag.CURRENT and branch != "main"):
        return f"{GITHUB_BASE}{branch}/{_build_path(source, fmt)}"

    # Return the latest published version
    elif release is ReleaseTag.CURRENT:
        release = do_request(f"{GITHUB_API_BASE}releases/latest")
        return tag_to_commit(release['tag_name'])

    # Return a specific tag
    else:
        return tag_to_commit(release)


class ModelFile:
    class ModelLoc:
        def __init__(self, model: Source, fmt: Format) -> str:
            self._model = model
            self._format = fmt

        def __str__(self):
            return f"{self._model.value}.{self._format.value}"

        def __repr__(self):
            return str(self)

        @property
        def url(self) -> str:
            return URL_FOR(self._model, self._format)

        @property
        def file(self) -> str:
            return LOCAL_PATH_FOR(self._model, self._format)

        def github_loc(self, tag: Optional[str] = None, branch: Optional[str] = None, release: ReleaseTag = None) -> str:
            if not tag and not branch and not release:
                return GITHUB_IO_PATH_FOR(self._model, self._format)
            if tag:
                return GITHUB_PATH_FOR(self._model, self._format, tag, branch or "main")
            else:
                return GITHUB_PATH_FOR(self._model, self._format, release or ReleaseTag.CURRENT, branch or "main")

    def __init__(self, model: Source) -> None:
        self._model = model

    def __str__(self):
        return self._model.value

    def __repr__(self):
        return str(self)

    @property
    def yaml(self) -> ModelLoc:
        return ModelFile.ModelLoc(self._model, Format.YAML)

    @property
    def graphql(self) -> ModelLoc:
        return ModelFile.ModelLoc(self._model, Format.GRAPHQL)

    @property
    def html(self) -> ModelLoc:
        return ModelFile.ModelLoc(self._model, Format.HTML)

    @property
    def json(self) -> ModelLoc:
        return ModelFile.ModelLoc(self._model, Format.JSON)

    @property
    def jsonld(self) -> ModelLoc:
        return ModelFile.ModelLoc(self._model, Format.JSONLD)

    @property
    def jsonschema(self) -> ModelLoc:
        return ModelFile.ModelLoc(self._model, Format.JSON_SCHEMA)

    @property
    def model_jsonld(self) -> ModelLoc:
        return ModelFile.ModelLoc(self._model, Format.NATIVE_JSONLD)

    @property
    def model_rdf(self) -> ModelLoc:
        return ModelFile.ModelLoc(self._model, Format.NATIVE_RDF)

    @property
    def model_shexc(self) -> ModelLoc:
        return ModelFile.ModelLoc(self._model, Format.NATIVE_SHEXC)

    @property
    def model_shexj(self) -> ModelLoc:
        return ModelFile.ModelLoc(self._model, Format.NATIVE_SHEXJ)

    @property
    def owl(self) -> ModelLoc:
        return ModelFile.ModelLoc(self._model, Format.OWL)

    @property
    def python(self) -> ModelLoc:
        return ModelFile.ModelLoc(self._model, Format.PYTHON)

    @property
    def rdf(self) -> ModelLoc:
        return ModelFile.ModelLoc(self._model, Format.RDF)

    @property
    def shexc(self) -> ModelLoc:
        return ModelFile.ModelLoc(self._model, Format.SHEXC)

    @property
    def shexj(self) -> ModelLoc:
        return ModelFile.ModelLoc(self._model, Format.SHEXJ)

    @property
    def yaml(self) -> ModelLoc:
        return ModelFile.ModelLoc(self._model, Format.YAML)


meta = ModelFile(Source.META)
types = ModelFile(Source.TYPES)
annotations = ModelFile(Source.ANNOTATIONS)
extensions = ModelFile(Source.EXTENSIONS)
mappings = ModelFile(Source.MAPPINGS)



import os
import sys
from logging import warning
from types import ModuleType


def file_text(txt_or_fname: str) -> str:
    """
    Determine whether text_or_fname is a file name or a string and, if a file name, read it
    :param txt_or_fname:
    :return:
    """
    if len(txt_or_fname) > 4 and '\n' not in txt_or_fname:
        with open(txt_or_fname) as ef:
            return ef.read()
    return txt_or_fname


def compile_python(text_or_fn: str, package_path: str = None) -> ModuleType:
    """
    Compile the text or file and return the resulting module
    @param text_or_fn: Python text or file name that references python file
    @param package_path: Root package path.  If omitted and we've got a python file, the package is the containing
    directory
    @return: Compiled module
    """
    python_txt = file_text(text_or_fn)
    if package_path is None and python_txt != text_or_fn:
        package_path = text_or_fn
    spec = compile(python_txt, 'test', 'exec')
    module = ModuleType('test')
    if package_path:
        package_path_abs = os.path.join(os.getcwd(), package_path)
        # We have to calculate the path to expected path relative to the current working directory
        for path in sys.path:
            if package_path.startswith(path):
                path_from_tests_parent = os.path.relpath(package_path, path)
                break
            if package_path_abs.startswith(path):
                path_from_tests_parent = os.path.relpath(package_path_abs, path)
                break
        else:
            warning(f"There is no established path to {package_path} - compile_python may or may not work")
            path_from_tests_parent = os.path.relpath(package_path, os.path.join(os.getcwd(), '..'))
        module.__package__ = os.path.dirname(os.path.relpath(path_from_tests_parent, os.getcwd())).replace(os.path.sep, '.')
    sys.modules[module.__name__] = module
    exec(spec, module.__dict__)
    return module



from dataclasses import fields
from typing import Union, Optional, Type

from linkml_runtime.utils.metamodelcore import Curie
from linkml_runtime.utils.yamlutils import YAMLRoot


class EnumDefinitionMeta(type):
    def __init__(cls, *args, **kwargs):
        super().__init__(*args, **kwargs)
        cls._addvals()

    def __getitem__(cls, item):
        return cls.__dict__[item]

    def __setitem__(cls, key, value):
        if key in cls.__dict__:
            raise ValueError(f"{cls.__name__} - {key} already assigned")
        cls.__dict__[key] = value

    def __setattr__(self, key, value):
        from linkml_runtime.linkml_model.meta import PermissibleValue
        if self._defn.code_set and isinstance(value, PermissibleValue) and value.meaning:
            print(f"Validating {value.meaning} against {self._defn.code_set}")
        super().__setattr__(key, value)

    def __contains__(cls, item) -> bool:
        return item in cls.__dict__


def isinstance_dt(cls: Type, inst: str) -> bool:
    """ Duck typing isinstance to prevent recursion errors """
    return inst in [c.__name__ for c in type(cls).mro()]

class EnumDefinitionImpl(YAMLRoot, metaclass=EnumDefinitionMeta):
    _defn: "EnumDefinition" = None         # Overridden by implementation

    def __init__(self, code: Union[str, Curie]) -> None:
        if isinstance_dt(code, "PermissibleValue"):
            key = code.text
        elif isinstance(code, Curie):
            key = str(code)
        else:
            key = code

        if key not in self.__class__ and self._defn.code_set:
            code = self._lookup(key)
            if code:
                self.__class__[key] = code
                self._code = code
        elif key not in self.__class__:
            raise ValueError(f"Unknown {self.__class__.__name__} enumeration code: {key}")
        elif isinstance_dt(code, "PermissibleValue"):
            if getattr(self, 'code', None):
                if self._code != code:
                    raise ValueError(f"Enumeration: {self.__class__.__name__} - "
                                     f"Cannot change an existing permissible value entry for {code}")
            else:
                self._code = code
        else:
            self._code = self.__class__[key]

    def _lookup(self, key: str) -> Optional["PermissibleValue"]:
        """
        Hook to look up key in the appropriate code system
        @param key: URI or string in Curie form (TBD)
        @return: Permissible value rendering if key is valid
        """
        return None

    # WARNING: any non "_" prefix pollutes the EnumDefinition namespace.  These CAN be overridden, but be aware that
    # the value "code", "meaning" "uri", or "curie" as actual codes will mean that one will need to use "_code" and
    # direct access to get at the real values
    @property
    def code(self):
        return self._code

    @code.setter
    def code(self, val):
        self._code = val

    @property
    def meaning(self):
        return self._code.meaning

    @property
    def uri(self):
        return self._code.meaning

    @property
    def curie(self):
        return "Curie for: " + self._code.meaning

    @classmethod
    def _addvals(cls):
        """ Override this to add non-python compatible values """
        pass

    def __str__(self) -> str:
        """ The string representation of an enumerated value should be the code representing this value."""
        return self._code.text

    def __repr__(self) -> str:
        rlist = [(f.name, getattr(self._code, f.name)) for f in fields(self._code)]
        return '(' + ', '.join([f"{f[0]}={repr(f[1])}" for f in rlist if f[1]]) + ')'



from typing import List, Union

from linkml_runtime.utils.schemaview import SchemaView, CLASS_NAME
from linkml_runtime.linkml_model.meta import SchemaDefinition, ClassDefinition

CLASS_NAME_OR_LIST = Union[CLASS_NAME, List[CLASS_NAME]]

def roll_up(sv: SchemaView, classes: CLASS_NAME_OR_LIST = None, mixins=True, is_a=True, delete=True) -> None:
    """
    rolls slots up to a set of ancestor classes

    :param sv: view over source schema
    :param classes: all slots will be rolled up to these classes
    :param mixins: include mixins (default is True)
    :param is_a: include is_a parents (default is True)
    :return:
    """
    if classes is None:
        classes = sv.class_roots()
    if not isinstance(classes, list):
        classes = [classes]
    dels = set()
    for cn in classes:
        c = sv.get_class(cn)
        slots = []
        for d in sv.class_descendants(cn, reflexive=False, mixins=mixins, is_a=is_a):
            for sn in sv.class_slots(d):
                s = sv.induced_slot(sn, class_name=d)
                if sn not in c.slots:
                    c.slots.append(sn)
                # if slot was declared as an attribute it may not be in the set of slots
                if sn not in sv.schema.slots:
                    # TODO: determine behavior for merging
                    sv.add_slot(s)
            if delete:
                dels.add(d)
    for d in dels:
        sv.delete_class(d)
    sv.set_modified()

def roll_down(sv: SchemaView, classes: List[CLASS_NAME] = None, mixins=True, is_a=True, delete=True) -> None:
    """
    rolls down to a set of descendant classes

    :param sv:
    :param classes:
    :param mixins: include mixins (default is True)
    :param is_a: include is_a parents (default is True)
    :return:
    """
    if classes is None:
        classes = sv.class_leaves()
    dels = set()
    for cn in classes:
        c = sv.get_class(cn)
        slots = []
        for d in sv.class_ancestors(cn, reflexive=False, mixins=mixins, is_a=is_a):
            for sn in sv.class_slots(d):
                s = sv.induced_slot(sn, class_name=d)
                if sn not in c.slots:
                    c.slots.append(sn)
                # if slot was declared as an attribute it may not be in the set of slots
                if sn not in sv.schema.slots:
                    # TODO: determine behavior for merging
                    sv.add_slot(s)
            if delete:
                dels.add(d)
        # when we roll down with deletion, parent classes disappear,
        # cascading
        if delete and False:
            if mixins:
                c.mixins = []
            if is_a and c.is_a is not None:
                del c.is_a
    for d in dels:
        d_cls = sv.get_class(d)
        sv.delete_class(d)
    sv.set_modified()






import sys

# The purpose behind the set of changes below is to change two signatures in the dataclasses
# generated code.  The first is to add the **kwargs at the end:
#     def __init__(self, a1, a2, ..., p1=k1, p2=k2, ...pn=kn, **kwargs)
#
# this allows any named parameter that isn't in the list to pass through the dataclasses init section.
#
# The second change is to the __post_init__ itself:
#     def __post_init__(self, ...,  **kwargs)
#
# This gives control over error reporting and the like back to the generated python, which
# allows it to give line numbers and other things.
#
# NOTE: we copy the _init_fn(...) signature from the python supplied dataclasses.py and change
#       the two lines.

if sys.version_info < (3, 7, 0):
    raise NotImplementedError("LinkML requires Python 3.7 or later to run")
elif sys.version_info >= (3, 10, 0):
    # dataclasses takes 8 arguments
    # Taken from 3.10.1
    from dataclasses import MISSING, _HAS_DEFAULT_FACTORY, _field_init, _FIELD_INITVAR, _POST_INIT_NAME, _init_param, \
        _create_fn

    def dataclasses_init_fn_with_kwargs(fields, std_fields, kw_only_fields, frozen, has_post_init,
                                        self_name, globals, slots):
        # fields contains both real fields and InitVar pseudo-fields.

        # Make sure we don't have fields without defaults following fields
        # with defaults.  This actually would be caught when exec-ing the
        # function source code, but catching it here gives a better error
        # message, and future-proofs us in case we build up the function
        # using ast.

        seen_default = False
        for f in std_fields:
            # Only consider the non-kw-only fields in the __init__ call.
            if f.init:
                if not (f.default is MISSING and f.default_factory is MISSING):
                    seen_default = True
                elif seen_default:
                    raise TypeError(f'non-default argument {f.name!r} '
                                    'follows default argument')

        locals = {f'_type_{f.name}': f.type for f in fields}
        locals.update({
            'MISSING': MISSING,
            '_HAS_DEFAULT_FACTORY': _HAS_DEFAULT_FACTORY,
        })

        body_lines = []
        for f in fields:
            line = _field_init(f, frozen, locals, self_name, slots)
            # line is None means that this field doesn't require
            # initialization (it's a pseudo-field).  Just skip it.
            if line:
                body_lines.append(line)

        # Does this class have a post-init function?
        if has_post_init:
            params_str = ','.join(f.name for f in fields
                                  if f._field_type is _FIELD_INITVAR)
            body_lines.append(f'{self_name}.{_POST_INIT_NAME}({params_str}{", " if params_str else ""} **kwargs)')

        # If no body lines, use 'pass'.
        if not body_lines:
            body_lines = ['pass']

        _init_params = [_init_param(f) for f in std_fields]
        if kw_only_fields:
            # Add the keyword-only args.  Because the * can only be added if
            # there's at least one keyword-only arg, there needs to be a test here
            # (instead of just concatenting the lists together).
            _init_params += ['*']
            _init_params += [_init_param(f) for f in kw_only_fields]
        return _create_fn('__init__',
                          [self_name] + _init_params + ["**kwargs"],
                          body_lines,
                          locals=locals,
                          globals=globals,
                          return_type=None)
elif sys.version_info.minor == 7:
    # 5 Arguments
    # Taken from 3.7.9
    from dataclasses import MISSING, _HAS_DEFAULT_FACTORY, _field_init, _FIELD_INITVAR, _POST_INIT_NAME, _init_param, \
        _create_fn

    def dataclasses_init_fn_with_kwargs(fields, frozen, has_post_init, self_name, globals):
        # fields contains both real fields and InitVar pseudo-fields.

        # Make sure we don't have fields without defaults following fields
        # with defaults.  This actually would be caught when exec-ing the
        # function source code, but catching it here gives a better error
        # message, and future-proofs us in case we build up the function
        # using ast.
        seen_default = False
        for f in fields:
            # Only consider fields in the __init__ call.
            if f.init:
                if not (f.default is MISSING and f.default_factory is MISSING):
                    seen_default = True
                elif seen_default:
                    raise TypeError(f'non-default argument {f.name!r} '
                                    'follows default argument')

        locals = {f'_type_{f.name}': f.type for f in fields}
        locals.update({
            'MISSING': MISSING,
            '_HAS_DEFAULT_FACTORY': _HAS_DEFAULT_FACTORY,
        })

        body_lines = []
        for f in fields:
            line = _field_init(f, frozen, locals, self_name)
            # line is None means that this field doesn't require
            # initialization (it's a pseudo-field).  Just skip it.
            if line:
                body_lines.append(line)

        # Does this class have a post-init function?
        if has_post_init:
            params_str = ','.join(f.name for f in fields
                                  if f._field_type is _FIELD_INITVAR)
            # This line is changed
            body_lines.append(f'{self_name}.{_POST_INIT_NAME}({params_str}{", " if params_str else ""} **kwargs)')

        # If no body lines, use 'pass'.
        if not body_lines:
            body_lines = ['pass']

        return _create_fn('__init__',
                          # This line is changed
                          [self_name] + [_init_param(f) for f in fields if f.init] + ["**kwargs"],
                          body_lines,
                          locals=locals,
                          globals=globals,
                          return_type=None)
elif sys.version_info.minor == 8 and sys.version_info.micro == 0:
    # Taken from 3.8.0
    # Dataclass takes 4 arguments
    from dataclasses import MISSING, _HAS_DEFAULT_FACTORY, _field_init, _FIELD_INITVAR, _POST_INIT_NAME, _init_param, \
        _create_fn

    def dataclasses_init_fn_with_kwargs(fields, frozen, has_post_init, self_name):
        # fields contains both real fields and InitVar pseudo-fields.

        # Make sure we don't have fields without defaults following fields
        # with defaults.  This actually would be caught when exec-ing the
        # function source code, but catching it here gives a better error
        # message, and future-proofs us in case we build up the function
        # using ast.
        seen_default = False
        for f in fields:
            # Only consider fields in the __init__ call.
            if f.init:
                if not (f.default is MISSING and f.default_factory is MISSING):
                    seen_default = True
                elif seen_default:
                    raise TypeError(f'non-default argument {f.name!r} '
                                    'follows default argument')

        globals = {'MISSING': MISSING,
                   '_HAS_DEFAULT_FACTORY': _HAS_DEFAULT_FACTORY}

        body_lines = []
        for f in fields:
            line = _field_init(f, frozen, globals, self_name)
            # line is None means that this field doesn't require
            # initialization (it's a pseudo-field).  Just skip it.
            if line:
                body_lines.append(line)

        # Does this class have a post-init function?
        if has_post_init:
            params_str = ','.join(f.name for f in fields
                                  if f._field_type is _FIELD_INITVAR)
            # This line is changed
            body_lines.append(f'{self_name}.{_POST_INIT_NAME}({params_str}{", " if params_str else ""} **kwargs)')

        # If no body lines, use 'pass'.
        if not body_lines:
            body_lines = ['pass']

        locals = {f'_type_{f.name}': f.type for f in fields}
        return _create_fn('__init__',
                          # This line is changed
                          [self_name] + [_init_param(f) for f in fields if f.init] + ["**kwargs"],
                          body_lines,
                          locals=locals,
                          globals=globals,
                          return_type=None)
elif sys.version_info.minor == 8 and sys.version_info.micro > 0:
    # Taken from 3.8.12
    # Dataclass takes 5 arguments
    from dataclasses import MISSING, _HAS_DEFAULT_FACTORY, _field_init, _FIELD_INITVAR, _POST_INIT_NAME, _init_param, \
        _create_fn

    def dataclasses_init_fn_with_kwargs(fields, frozen, has_post_init, self_name, globals):
        # fields contains both real fields and InitVar pseudo-fields.

        # Make sure we don't have fields without defaults following fields
        # with defaults.  This actually would be caught when exec-ing the
        # function source code, but catching it here gives a better error
        # message, and future-proofs us in case we build up the function
        # using ast.
        seen_default = False
        for f in fields:
            # Only consider fields in the __init__ call.
            if f.init:
                if not (f.default is MISSING and f.default_factory is MISSING):
                    seen_default = True
                elif seen_default:
                    raise TypeError(f'non-default argument {f.name!r} '
                                    'follows default argument')

        locals = {f'_type_{f.name}': f.type for f in fields}
        locals.update({
            'MISSING': MISSING,
            '_HAS_DEFAULT_FACTORY': _HAS_DEFAULT_FACTORY,
        })

        body_lines = []
        for f in fields:
            line = _field_init(f, frozen, locals, self_name)
            # line is None means that this field doesn't require
            # initialization (it's a pseudo-field).  Just skip it.
            if line:
                body_lines.append(line)

        # Does this class have a post-init function?
        if has_post_init:
            params_str = ','.join(f.name for f in fields
                                  if f._field_type is _FIELD_INITVAR)
            # This line is changed
            body_lines.append(f'{self_name}.{_POST_INIT_NAME}({params_str}{", " if params_str else ""} **kwargs)')


        # If no body lines, use 'pass'.
        if not body_lines:
            body_lines = ['pass']

        return _create_fn('__init__',
                          # This line is changed
                          [self_name] + [_init_param(f) for f in fields if f.init] + ["**kwargs"],
                          body_lines,
                          locals=locals,
                          globals=globals,
                          return_type=None)
elif sys.version_info.minor == 9:
    # Taken from 3.9.10
    # Dataclass takes 5 arguments
    from dataclasses import MISSING, _HAS_DEFAULT_FACTORY, _field_init, _FIELD_INITVAR, _POST_INIT_NAME, _init_param, \
        _create_fn

    def dataclasses_init_fn_with_kwargs(fields, frozen, has_post_init, self_name, globals):
        # fields contains both real fields and InitVar pseudo-fields.

        # Make sure we don't have fields without defaults following fields
        # with defaults.  This actually would be caught when exec-ing the
        # function source code, but catching it here gives a better error
        # message, and future-proofs us in case we build up the function
        # using ast.
        seen_default = False
        for f in fields:
            # Only consider fields in the __init__ call.
            if f.init:
                if not (f.default is MISSING and f.default_factory is MISSING):
                    seen_default = True
                elif seen_default:
                    raise TypeError(f'non-default argument {f.name!r} '
                                    'follows default argument')

        locals = {f'_type_{f.name}': f.type for f in fields}
        locals.update({
            'MISSING': MISSING,
            '_HAS_DEFAULT_FACTORY': _HAS_DEFAULT_FACTORY,
        })

        body_lines = []
        for f in fields:
            line = _field_init(f, frozen, locals, self_name)
            # line is None means that this field doesn't require
            # initialization (it's a pseudo-field).  Just skip it.
            if line:
                body_lines.append(line)

        # Does this class have a post-init function?
        if has_post_init:
            params_str = ','.join(f.name for f in fields
                                  if f._field_type is _FIELD_INITVAR)
            # This line is changed
            body_lines.append(f'{self_name}.{_POST_INIT_NAME}({params_str}{", " if params_str else ""} **kwargs)')


        # If no body lines, use 'pass'.
        if not body_lines:
            body_lines = ['pass']

        return _create_fn('__init__',
                          # This line is changed
                          [self_name] + [_init_param(f) for f in fields if f.init] + ["**kwargs"],
                          body_lines,
                          locals=locals,
                          globals=globals,
                          return_type=None)
else:
    raise NotImplementedError(f"Unrecognized python version - {sys.version}")


import logging
from copy import copy
from dataclasses import dataclass
from typing import Optional, Any, Dict, Type, Union, TextIO, List, Tuple, Set

from hbreader import FileInfo
from rdflib import Graph, URIRef
from rdflib.term import Node, BNode, Literal
from rdflib.namespace import RDF

from linkml_runtime import MappingError, DataNotFoundError
from linkml_runtime.linkml_model import ClassDefinitionName, TypeDefinition, EnumDefinition, ClassDefinition
from linkml_runtime.loaders.loader_root import Loader
from linkml_runtime.utils.formatutils import underscore
from linkml_runtime.utils.schemaview import SchemaView, SlotDefinition
from linkml_runtime.utils.yamlutils import YAMLRoot

VALID_SUBJECT = Union[URIRef, BNode]
ANYDICT = Dict[str, Any]

@dataclass
class Pointer:
    obj: str

class RDFLibLoader(Loader):
    """
    Loads objects from rdflib Graphs into the python target_class structure

    Note: this is a more complete replacement for rdf_loader
    """
    def from_rdf_graph(self, graph: Graph, schemaview: SchemaView, target_class: Type[YAMLRoot],
                       prefix_map: Dict[str, str] = None,
                       cast_literals: bool = True,
                       allow_unprocessed_triples: bool = True,
                       ignore_unmapped_predicates: bool = False) -> List[YAMLRoot]:
        """
        Loads objects from graph into lists of the python target_class structure,
        recursively walking RDF graph from instances of target_class.

        :param graph: rdflib Graph that holds instances of target_class
        :param schemaview: schema to which graph conforms
        :param target_class: class which root nodes should instantiate
        :param prefix_map: additional prefix mappings for data objects
        :param ignore_unmapped_predicates: if True then a predicate that has no mapping to a slot does not raise an error
        :return: all instances of target class type
        """
        namespaces = schemaview.namespaces()
        uri_to_class_map = {}
        for cn, c in schemaview.all_classes().items():
            uri = schemaview.get_uri(c, expand=True)
            if uri in uri_to_class_map:
                c2 = uri_to_class_map[uri]
                if c2.name in schemaview.class_ancestors(cn):
                    continue
                else:
                    logging.error(f'Inconsistent URI to class map: {uri} -> {c2.name}, {c.name}')
            uri_to_class_map[uri] = c
        # data prefix map: supplements or overrides existing schema prefix map
        if prefix_map:
            for k, v in prefix_map.items():
                namespaces[k] = v
                graph.namespace_manager.bind(k, URIRef(v))
        # Step 1: Create stub root dict-objects
        target_class_uriref: URIRef = target_class.class_class_uri
        root_dicts: List[ANYDICT] = []
        root_subjects: List[VALID_SUBJECT] = list(graph.subjects(RDF.type, target_class_uriref))
        logging.debug(f'ROOTS = {root_subjects}')
        # Step 2: walk RDF graph starting from root subjects, constructing dict tree
        node_tuples_to_visit: List[Tuple[VALID_SUBJECT, ClassDefinitionName]]  ## nodes and their type still to visit
        node_tuples_to_visit = [(subject, target_class.class_name) for subject in root_subjects]
        uri_to_slot: Dict[str, SlotDefinition]  ## lookup table for RDF predicates -> slots
        uri_to_slot = {URIRef(schemaview.get_uri(s, expand=True)): s for s in schemaview.all_slots().values()}
        processed: Set[VALID_SUBJECT] = set()  ## track nodes already visited, or already scheduled
        for n, _ in node_tuples_to_visit:
            processed.add(n)
        obj_map: Dict[VALID_SUBJECT, ANYDICT] = {}  ## map from an RDF node to its dict representation
        unmapped_predicates = set()
        processed_triples: Set[Tuple] = set()
        while len(node_tuples_to_visit) > 0:
            subject, subject_class = node_tuples_to_visit.pop()
            processed.add(subject)
            dict_obj = self._get_id_dict(subject, schemaview, subject_class)
            if subject in root_subjects:
                root_dicts.append(dict_obj)
            obj_map[subject] = dict_obj
            type_designator_slot = schemaview.get_type_designator_slot(subject_class)
            if type_designator_slot:
                td_iri = schemaview.get_uri(type_designator_slot, expand=True)
                type_vals = list(graph.objects(subject, URIRef(td_iri)))
                if len(type_vals) > 0:
                    type_classes = [uri_to_class_map[str(x)] for x in type_vals]
                    if len(type_classes) > 1:
                        raise ValueError(f'Ambiguous types for {subject} == {type_classes}')
                    logging.info(f'Replacing {subject_class} with {type_classes}')
                    subject_class = type_classes[0].name
            # process all triples for this node
            for (_, p, o) in graph.triples((subject, None, None)):
                processed_triples.add((subject,p,o))
                logging.debug(f' Processing triple {subject} {p} {o}, subject type = {subject_class}')
                if p == RDF.type:
                    logging.debug(f'Ignoring RDF.type for {subject} {o}, we automatically infer this from {subject_class}')
                elif p not in uri_to_slot:
                    if ignore_unmapped_predicates:
                        unmapped_predicates.add(p)
                    else:
                        raise MappingError(f'No pred for {p} {type(p)}')
                else:
                    slot = schemaview.induced_slot(uri_to_slot[p].name, subject_class)
                    range_applicable_elements = schemaview.slot_applicable_range_elements(slot)
                    is_inlined = schemaview.is_inlined(slot)
                    slot_name = underscore(slot.name)
                    if isinstance(o, Literal):
                        if EnumDefinition.class_name in range_applicable_elements:
                            logging.debug(f'Assuming no meaning assigned for value {o} for Enum {slot.range}')
                        elif TypeDefinition.class_name not in range_applicable_elements:
                            raise ValueError(f'Cannot map Literal {o} to a slot {slot.name} whose range {slot.range} is not a type;')
                        v = o.value
                    elif isinstance(o, BNode):
                        if not is_inlined:
                            logging.error(f'blank nodes should be inlined; {slot_name}={o} in {subject}')
                        v = Pointer(o)
                    else:
                        if ClassDefinition.class_name in range_applicable_elements:
                            v = namespaces.curie_for(o)
                            if v is None:
                                logging.debug(f'No CURIE for {p}={o} in {subject} [{subject_class}]')
                                v = str(o)
                        elif EnumDefinition.class_name in range_applicable_elements:
                            # if a PV has a meaning URI declared, map this
                            # back to a text representation
                            v = namespaces.curie_for(o)
                            e = schemaview.get_enum(slot.range)
                            for pv in e.permissible_values.values():
                                if v == pv.meaning or str(o) == pv.meaning:
                                    v = pv.text
                                    break
                        elif TypeDefinition.class_name in range_applicable_elements:
                            if cast_literals:
                                v = namespaces.curie_for(o)
                                if v is None:
                                    v = str(o)
                                logging.debug(f'Casting {o} to string')
                            else:
                                raise ValueError(f'Expected literal value ({range_applicable_elements}) for {slot_name}={o}')
                        if is_inlined:
                            # the object of the triple may not yet be processed;
                            # we store a pointer to o, and then replace this later
                            v = Pointer(o)
                    if slot.multivalued:
                        if slot_name not in dict_obj:
                            dict_obj[slot_name] = []
                        dict_obj[slot_name].append(v)
                    else:
                        dict_obj[slot_name] = v
                    if o not in processed:
                        # if o instantiates a class, add to list of nodes to be visited.
                        # force type based on range constraint
                        if slot.range in schemaview.all_classes():
                            node_tuples_to_visit.append((o, ClassDefinitionName(slot.range)))
        if unmapped_predicates:
            logging.info(f'Unmapped predicated: {unmapped_predicates}')
        unprocessed_triples = set(graph.triples((None, None, None))) - processed_triples
        logging.info(f'Triple processed = {len(processed_triples)}, unprocessed = {len(unprocessed_triples)}')
        if len(unprocessed_triples) > 0:
            if not allow_unprocessed_triples:
                for t in unprocessed_triples:
                    logging.warning(f'  Unprocessed: {t}')
                raise ValueError(f'Unprocessed triples: {len(unprocessed_triples)}')
        # Step 2: replace inline pointers with object dicts
        def repl(v):
            if isinstance(v, Pointer):
                v2 = obj_map[v.obj]
                if v2 is None:
                    raise Exception(f'No mapping for pointer {v}')
                return v2
            else:
                return v
        objs_to_visit: List[ANYDICT] = copy(root_dicts)
        while len(objs_to_visit) > 0:
            obj = objs_to_visit.pop()
            logging.debug(f'Replacing pointers for  {obj}')
            for k, v in obj.items():
                if v is None:
                    continue
                if isinstance(v, list):
                    v = [repl(v1) for v1 in v if v1 is not None]
                    for v1 in v:
                        if isinstance(v1, dict):
                            objs_to_visit.append(v1)
                else:
                    v = repl(v)
                    if isinstance(v, dict):
                        objs_to_visit.append(v)
                obj[k] = v
        # Final step: translate dicts into instances of target_class
        return [target_class(**x) for x in root_dicts]

    def _get_id_dict(self, node: VALID_SUBJECT, schemaview: SchemaView, cn: ClassDefinitionName) -> ANYDICT:
        id_slot = schemaview.get_identifier_slot(cn)
        if not isinstance(node, BNode):
            id_val = schemaview.namespaces().curie_for(node)
            if id_val == None:
                id_val = str(node)
            return {id_slot.name: id_val}
        else:
            if id_slot is not None:
                raise Exception(f'Unexpected blank node {node}, type {cn} expects {id_slot.name} identifier')
            return {}


    def load(self, source: Union[str, TextIO, Graph], target_class: Type[YAMLRoot], *,
             schemaview: SchemaView = None,
             prefix_map: Dict[str, str] = None,
             fmt: Optional[str] = 'turtle',
             metadata: Optional[FileInfo] = None,
             **kwargs) -> YAMLRoot:
        """
        Load the RDF in source into the python target_class structure

        The assumption of all loaders is that the source contains exactly one instance of the
        target class. To load from graphs with multiple instances, use from_rdf_graph

        :param source: RDF data source. Can be a file name, an open handle or an existing graph
        :param target_class: LinkML class to load the RDF into
        :param schemaview: view over schema to guide instantiation
        :param prefix_map: map of prefixes used in data
        :param fmt: format of source if it isn't an existing Graph
        :param metadata: source information. Used by some loaders to record where information came from
        :param kwargs: additional arguments passed to from_rdf_graph
        :return: Instance of target_class
        """
        if isinstance(source, Graph):
            g = source
        else:
            g = Graph()
            if '\n' in source:
                g.parse(data=source, format=fmt)
            else:
                g.parse(source, format=fmt)
        objs = self.from_rdf_graph(g, schemaview=schemaview, target_class=target_class, prefix_map=prefix_map, **kwargs)
        if len(objs) != 1:
            raise DataNotFoundError(f'Got {len(objs)} of type {target_class} from source, expected exactly 1')
        return objs[0]

    def loads(self, source: str, **kwargs) -> YAMLRoot:
        return self.load(source, **kwargs)

    def load_any(self, source: str, **kwargs) -> Union[YAMLRoot, List[YAMLRoot]]:
        return self.load(source, **kwargs)





import json
from decimal import Decimal
from typing import Dict

from deprecated.classic import deprecated

from linkml_runtime.dumpers.dumper_root import Dumper
from linkml_runtime.utils import formatutils
from linkml_runtime.utils.context_utils import CONTEXTS_PARAM_TYPE
from linkml_runtime.utils.formatutils import remove_empty_items
from linkml_runtime.utils.yamlutils import YAMLRoot, as_json_object
from jsonasobj2 import JsonObj

class JSONDumper(Dumper):

    def dump(self, element: YAMLRoot, to_file: str, contexts: CONTEXTS_PARAM_TYPE = None, **kwargs) -> None:
        """
        Write element as json to to_file
        :param element: LinkML object to be serialized as YAML
        :param to_file: file to write to
        :param contexts: JSON-LD context(s) in the form of:
            * file name
            * URL
            * JSON String
            * dict
            * JSON Object
            * A list containing elements of any type named above
        """
        super().dump(element, to_file, contexts=contexts, **kwargs)

    def dumps(self, element: YAMLRoot, contexts: CONTEXTS_PARAM_TYPE = None, inject_type=True) -> str:
        """
        Return element as a JSON or a JSON-LD string
        :param element: LinkML object to be emitted
        :param contexts: JSON-LD context(s) in the form of:
            * file name
            * URL
            * JSON String
            * dict
            * JSON Object
            * A list containing elements of any type named above
        :param inject_type: if True (default), add a @type at the top level
        :return: JSON Object representing the element
        """
        def default(o):
            if isinstance(o, YAMLRoot):
                return remove_empty_items(o, hide_protected_keys=True)
            elif isinstance(o, Decimal):
                # https://stackoverflow.com/questions/1960516/python-json-serialize-a-decimal-object
                return str(o)
            else:
                return json.JSONDecoder().decode(o)
        return json.dumps(as_json_object(element, contexts, inject_type=inject_type),
                          default=default,
                          ensure_ascii=False,
                          indent='  ')


    @staticmethod
    @deprecated("Use `utils/formatutils/remove_empty_items` instead")
    def remove_empty_items(obj: Dict) -> Dict:
        """
        Remove empty items from obj
        :param obj:
        :return: copy of dictionary with empty lists/dicts and Nones removed
        """
        return formatutils.remove_empty_items(obj, hide_protected_keys=True)

    def to_json_object(self, element: YAMLRoot, contexts: CONTEXTS_PARAM_TYPE = None, inject_type=True) -> JsonObj:
        """
        As dumps(), except returns a JsonObj, not a string

        :param element: LinkML object to be emitted
        :param contexts: JSON-LD context(s) in the form of:
            * file name
            * URL
            * JSON String
            * dict
            * JSON Object
            * A list containing elements of any type named above
        :param inject_type: if True (default), add a @type at the top level
        :return: JSON Object representing the element
        """
        return as_json_object(element, contexts, inject_type=inject_type)

    def to_dict(self, element: YAMLRoot, **kwargs) -> JsonObj:
        """
        As dumps(), except returns a JsonObj, not a string

        :param element: LinkML object to be emitted
        :param contexts: JSON-LD context(s) in the form of:
            * file name
            * URL
            * JSON String
            * dict
            * JSON Object
            * A list containing elements of any type named above
        :param inject_type: if True (default), add a @type at the top level
        :return: JSON Object representing the element
        """
        return json.loads(self.dumps(element, inject_type=False, **kwargs))


import json
from typing import Dict

from linkml_runtime.dumpers import json_dumper
from linkml_runtime.utils.yamlutils import YAMLRoot

def as_simple_dict(element: YAMLRoot) -> Dict:
    """
    Returns the representation of element as a python dictionary

    :param element: element to return
    :return: simple python dictionary
    """
    obj = json_dumper.dumps(element=element, inject_type=False)
    return json.loads(obj)



import io
import json
import logging
import re
from pathlib import Path
from typing import Dict, List, Tuple

from json_flattener import flatten_to_csv

from linkml_runtime.linkml_model import Element
from linkml_runtime.utils.schemaview import SchemaView
from linkml_runtime.dumpers import json_dumper, yaml_dumper
import click
import yaml

DEFAULT_DISPLAY_COLS = [
    'name',
    'is_a',
    'description',
    'owner'
]

schema_option = click.option(
    "-s", "--schema", help="Path to schema in LinkML yaml."
)
columns_option = click.option(
    "-c", "--columns",
    default=','.join(DEFAULT_DISPLAY_COLS),
    help="Comma separate list of columns to display."
)

@click.group()
@click.option("-v", "--verbose", count=True)
@click.option("-q", "--quiet")
def main(verbose: int, quiet: bool):
    """Main

    """
    if verbose >= 2:
        logging.basicConfig(level=logging.DEBUG)
    elif verbose == 1:
        logging.basicConfig(level=logging.INFO)
    else:
        logging.basicConfig(level=logging.WARNING)
    if quiet:
        logging.basicConfig(level=logging.ERROR)


@main.command()
@schema_option
@columns_option
@click.option('-t', '--element-type', help='Element type')
def list(schema, columns, element_type):
    """List elements in schema

    Example:
        list -s personinfo.yaml -c name,description -t class

    """
    schema_view = SchemaView(schema)
    logging.info(f'id={schema_view.schema.id}')
    logging.info(f'name={schema_view.schema.name}')
    enames = schema_view.all_element()
    elements = [schema_view.get_element(ename) for ename in enames]
    if element_type is not None:
        elements = [e for e in elements if element_type in type(e).class_name]
    _show_elements(elements, columns=columns)

@main.command()
@schema_option
@columns_option
@click.argument('class_names', nargs=-1)
def islot(schema, columns, class_names):
    """Show induced slots for a list of classes

    Example:
        schemaview islot -s personinfo.yaml Person

    """
    schema_view = SchemaView(schema)
    for cn in class_names:
        logging.info(f'Class: {cn}')
        islots = schema_view.class_induced_slots(cn)
        _show_elements(islots, columns=columns)

@main.command()
@schema_option
@click.option('--is-a/--no-is-a', default=True, help='Include is_a')
@click.option('--mixins/--no-mixins', default=True, help='Include mixins')
@click.argument('class_names', nargs=-1)
def ancs(schema, class_names, is_a, mixins):
    """Show ancestors for classes

    Example:
        schemaview ancs -s personinfo.yaml Person

    """
    schema_view = SchemaView(schema)
    for cn in class_names:
        logging.info(f'Class: {cn}')
        ancs = schema_view.class_ancestors(cn, is_a=is_a, mixins=mixins)
        for a in ancs:
            print(f'{cn}\t{a}')

@main.command()
@schema_option
@click.option('--is-a/--no-is-a', default=True, help='Include is_a')
@click.option('--mixins/--no-mixins', default=True, help='Include mixins')
@click.argument('class_names', nargs=-1)
def descs(schema, class_names, is_a, mixins):
    """Show descendants for classes

    Example:
        schemaview ancs -s personinfo.yaml NamedThing

    """
    schema_view = SchemaView(schema)
    for cn in class_names:
        logging.info(f'Class: {cn}')
        ds = schema_view.class_descendants(cn, is_a=is_a, mixins=mixins)
        for d in ds:
            print(f'{cn}\t{d}')

@main.command()
@schema_option
@click.argument('class_names', nargs=-1)
def delete(schema, class_names):
    """Deletes classes

    """
    schema_view = SchemaView(schema)
    for cn in class_names:
        logging.info(f'Class: {cn}')
        schema_view.delete_class(cn)
    print(yaml_dumper.dumps(schema_view.schema))


def _show_elements(elements: List[Element], columns=None, output = io.StringIO()) -> None:
    elements_j = json.loads(json_dumper.dumps(elements, inject_type=False))
    if columns is not None and columns != '' and columns != [] and columns != '%':
        if isinstance(columns, str):
            columns = columns.split(',')
        def filter_columns(row: dict):
            return {k: row.get(k, None) for k in columns}
        elements_j = [filter_columns(e) for e in elements_j]
    flatten_to_csv(elements_j, output)
    print(output.getvalue())

if __name__ == "__main__":
    main()



import logging
from abc import abstractmethod
from typing import Optional, Any, Dict

from rdflib import Graph, URIRef
from rdflib.term import Node, BNode, Literal
from rdflib.namespace import RDF


from linkml_runtime.dumpers.dumper_root import Dumper
from linkml_runtime.utils.schemaview import SchemaView, ElementName, PermissibleValue, PermissibleValueText
from linkml_runtime.utils.yamlutils import YAMLRoot


class RDFLibDumper(Dumper):
    """
    Dumps from elements (instances of a LinkML model) to an rdflib Graph

    Note: this should be used in place of rdf_loader for now

    This requires a SchemaView object

    """
    def as_rdf_graph(self, element: YAMLRoot, schemaview: SchemaView, prefix_map: Dict[str, str] = None) -> Graph:
        """
        Dumps from element to an rdflib Graph,
        following a schema

        :param element: element to represent in RDF
        :param schemaview:
        :param prefix_map:
        :return:
        """
        g = Graph()
        logging.debug(f'PREFIXMAP={prefix_map}')
        if prefix_map:
            for k, v in prefix_map.items():
                schemaview.namespaces()[k] = v
                g.namespace_manager.bind(k, URIRef(v))
        self.inject_triples(element, schemaview, g)
        return g

    def inject_triples(self, element: Any, schemaview: SchemaView, graph: Graph, target_type: ElementName = None) -> Node:
        """
        Inject triples from conversion of element into a Graph

        :param element: element to represent in RDF
        :param schemaview:
        :param graph:
        :param target_type:
        :return: root node as rdflib URIRef, BNode, or Literal
        """
        namespaces = schemaview.namespaces()
        slot_name_map = schemaview.slot_name_mappings()
        logging.debug(f'CONVERT: {element} // {type(element)} // {target_type}')
        if target_type in schemaview.all_enums():
            if isinstance(element, PermissibleValueText):
                e = schemaview.get_enum(target_type)
                element = e.permissible_values[element]
            else:
                element = element.code
            element: PermissibleValue
            if element.meaning is not None:
                return URIRef(schemaview.expand_curie(element.meaning))
            else:
                return Literal(element.text)
        if target_type in schemaview.all_types():
            t = schemaview.get_type(target_type)
            dt_uri = t.uri
            if dt_uri:
                if dt_uri == 'rdfs:Resource':
                    return URIRef(schemaview.expand_curie(element))
                elif dt_uri == 'xsd:string':
                    return Literal(element)
                else:
                    return Literal(element, datatype=namespaces.uri_for(dt_uri))
            else:
                logging.warning(f'No datatype specified for : {t.name}, using plain Literal')
                return Literal(element)
        element_vars = {k: v for k, v in vars(element).items() if not k.startswith('_')}
        if len(element_vars) == 0:
            return URIRef(schemaview.expand_curie(str(element)))
        element_type = type(element)
        cn = element_type.class_name
        id_slot = schemaview.get_identifier_slot(cn)
        if id_slot is not None:
            element_id = getattr(element, id_slot.name)
            logging.debug(f'ELEMENT_ID={element_id} // {id_slot.name}')
            element_uri = namespaces.uri_for(element_id)
        else:
            element_uri = BNode()
        type_added = False
        for k, v_or_list in element_vars.items():
            if isinstance(v_or_list, list):
                vs = v_or_list
            elif isinstance(v_or_list, dict):
                vs = v_or_list.values()
            else:
                vs = [v_or_list]
            for v in vs:
                if v is None:
                    continue
                if k in slot_name_map:
                    k = slot_name_map[k].name
                else:
                    logging.error(f'Slot {k} not in name map')
                slot = schemaview.induced_slot(k, cn)
                if not slot.identifier:
                    slot_uri = URIRef(schemaview.get_uri(slot, expand=True))
                    v_node = self.inject_triples(v, schemaview, graph, slot.range)
                    graph.add((element_uri, slot_uri, v_node))
                    if slot.designates_type:
                        type_added = True
        if not type_added:
            graph.add((element_uri, RDF.type, URIRef(schemaview.get_uri(cn, expand=True))))
        return element_uri

    def dump(self, element: YAMLRoot,
             to_file: str,
             schemaview: SchemaView = None,
             fmt: str = 'turtle', prefix_map: Dict[str, str] = None, **args) -> None:
        """
        Write element as rdf to to_file

        :param element: element to represent in RDF
        :param to_file:
        :param schemaview:
        :param fmt:
        :param prefix_map:
        :return:
        """
        super().dump(element, to_file, schemaview=schemaview, fmt=fmt, prefix_map=prefix_map)

    def dumps(self, element: YAMLRoot, schemaview: SchemaView = None,
              fmt: Optional[str] = 'turtle', prefix_map: Dict[str, str] = None) -> str:
        """
        Convert element into an RDF graph guided by the schema

        :param element:
        :param schemaview:
        :param fmt:
        :param prefix_map:
        :return: serialization of rdflib Graph containing element
        """
        return self.as_rdf_graph(element, schemaview, prefix_map=prefix_map).\
            serialize(format=fmt)




import logging
from dataclasses import dataclass
from typing import Set, List, Union

from linkml_runtime.utils.schemaview import SchemaView, CLASS_NAME
from linkml_runtime.linkml_model.meta import SchemaDefinition, ClassDefinition, SlotDefinition, Expression, \
    ClassExpression, ClassDefinitionName, ClassRule, AnonymousClassExpression, SlotExpression, SlotDefinitionName


class AtomicClassExpression:
    """
    aka literal (https://en.wikipedia.org/wiki/Literal_(mathematical_logic))
    """
    expression: ClassExpression
    negated: bool

    def __str__(self):
        if self.negated:
            return f'NOT {self.expression}'
        else:
            return str(self.expression)

class ClassExpressionConjunction:
    """
    conjunction of class literals
    """
    operands: AtomicClassExpression

    def __str__(self):
        return ' AND '.join([str(op) for op in self.operands])

@dataclass
class DisjunctiveNormalFormClassExpression:
    """
    A boolean combination of class expressions in Disjunctive Normal Form
    """
    operands: List[ClassExpressionConjunction]

    def __str__(self):
        return ' OR '.join([str(op) for op in self.operands])


def get_range_as_disjunction(slot: SlotExpression) ->  Set[ClassDefinitionName]:
    """
    translate the range of a slot as defined by both range expressions and direct
    named class ranges to a disjunctive expression

    :param slot:
    :return:
    """
    conjs = []
    if slot.any_of or slot.exactly_one_of:
        disj = set()
        for s in slot.any_of + slot.exactly_one_of:
            disj.update(get_range_as_disjunction(s))
        conjs.append(disj)
    if slot.range_expression:
        if isinstance(slot.range_expression, ClassExpression):
            conjs.append(get_disjunction(slot.range_expression))
        else:
            logging.warning(f'Expected range_expression for {slot.name} to be a class expression, not {type(slot.range_expression)}')
    if len(conjs) == 0:
        if slot.range:
            conjs.append({slot.range})
        else:
            logging.warning(f'No range for {slot}')
    if len(conjs) > 1:
        raise Exception(f'Cannot determine range disjunction for {slot}, got conjunctions: {conjs}')
    if len(conjs) == 0:
        return None
    else:
        return conjs[0]


def get_disjunction(cx: ClassExpression) -> Set[ClassDefinitionName]:
    disj = set()
    if cx.is_a:
        disj.add(cx.is_a)
    for arg in cx.any_of + cx.exactly_one_of:
        if isinstance(arg, ClassExpression):
            disj.update(get_disjunction(arg))
    return disj


def subclass_to_rules(view: SchemaView, child: ClassDefinitionName, parent: ClassDefinitionName,
                      type_designator_slot: SlotDefinitionName = None) -> List[ClassRule]:
    """
    rolls up child class to parent class, turning class-specific slot_usages into rules
    :param view:
    :param child: class to roll up to
    :param parent: class to roll up from
    :return:
    """
    child_slots = view.class_induced_slots(child)
    rule = ClassRule()
    if type_designator_slot is None:
        type_designators = [s.name for s in view.class_induced_slots(parent) if s.designates_type]
        if len(type_designators) == 1:
            type_designator_slot = type_designators[0]
        elif len(type_designators) > 1:
            raise Exception(f'Multiple type designatirs: {type_designators}')
        else:
            type_designator_slot = SlotDefinitionName('type')
    rule.preconditions = AnonymousClassExpression(slot_conditions=[SlotDefinition('type', equals_string=child)])
    rule.postconditions = AnonymousClassExpression(slot_conditions=child_slots)
    # ensure slots are declared for parent
    parent_slots = view.class_induced_slots(parent)
    parent_slot_names = [s.name for s in parent_slots]
    parent_cls = view.get_class(parent)
    for child_slot in child_slots:
        if child_slot.name not in parent_slot_names:
            parent_cls.slots.append(child_slot.name)
    if not rule_subsumed_by_class(view, rule=rule, cls=parent_cls):
        parent_cls.rules.append(rule)
    return [rule]


def rule_subsumed_by_class(view: SchemaView, rule, cls: ClassDefinition):
    induced_slots = view.class_induced_slots(cls.name)
    return False


def remove_redundant_rules(view: SchemaView, class_name: ClassDefinitionName):
    induced_slots = view.class_induced_slots(class_name)
    cls = view.get_class(class_name)
    redundant_rules = []
    for rule in cls.rules:
        if rule_subsumed_by_class(view, rule, cls):
            redundant_rules.append(rule)
    for rule in redundant_rules:
        cls.rules.remove(rule)
    return redundant_rules





